# -*- coding: utf-8 -*-
import itertools
import copy
from collections import OrderedDict
from types import SimpleNamespace
from keyword import iskeyword
import datetime
import importlib
import numpy as np
from pyyeti import locate


__all__ = [
    "_is_valid_identifier",
    "get_drfunc",
    "_merge_uf_reds",
    "_get_rpt_headers",
    "_get_numform",
    "get_marker_cycle",
    "_proc_filterval",
    "PrintCLAInfo",
    "freq3_augment",
    "maxmin",
    "nan_argmax",
    "nan_argmin",
    "nan_absmax",
    "extrema",
    #     'reorder',
    "_calc_covariance_sine_cosine",
    "PSD_consistent_rss",
]


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


def _is_valid_identifier(name):
    "Return True if `name` is valid Python identifier"
    return name.isidentifier() and not iskeyword(name)


def get_drfunc(filename, funcname, get_psd=False):
    """
    Get data recovery function(s)

    Parameters
    ----------
    filename : string
        Name of Python file containing data recovery function
    funcname : string
        Name of data recovery function
    get_psd : bool; optional
        If True, return a tuple of (`drfunc`, `psd_drfunc`).
        Otherwise, just return `drfunc`.

    Returns
    -------
    drfunc : function
        The data recovery function
    psd_drfunc : function or None; optional
        The PSD-specific data recovery function or None if no such
        function was defined; only returned if `get_psd` is True.
    """
    if _is_valid_identifier(funcname):
        # force a proper exception if file doesn't exist:
        with open(filename, "r") as _:
            pass
        spec = importlib.util.spec_from_file_location("has_drfuncs", filename)
        drmod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(drmod)
        func = getattr(drmod, funcname)
        if get_psd:
            psdfunc = getattr(drmod, funcname + "_psd", None)
            return func, psdfunc
        return func

    # build function and return it:
    strfunc = "def _func(sol, nas, Vars, se):\n    return " + funcname.strip()
    g = globals()
    exec(strfunc, g)
    if get_psd:
        return g["_func"], None
    return g["_func"]


def _merge_uf_reds(old, new, method="replace"):
    if method == "replace":
        merged = (n if n is not None else o for o, n in zip(old, new))
    elif method == "multiply":
        merged = (o * n if n is not None else o for o, n in zip(old, new))
    elif callable(method):
        merged = (method(o, n) if n is not None else o for o, n in zip(old, new))
    else:
        raise ValueError(
            '`method` value must be either "replace", '
            '"multiply", or a function (a callable)'
        )
    return tuple(merged)


def _get_rpt_headers(res=None, desc=None, uf_reds=None, units=None, misc=""):
    if res is not None:
        desc = res.drminfo.desc
        uf_reds = res.drminfo.uf_reds
        units = res.drminfo.units
    descline = f"Description: {desc}\n"
    if uf_reds is None:
        unceline = "Uncertainty: Not specified\n"
    else:
        unceline = (
            "Uncertainty: [Rigid, Elastic, Dynamic, Static] "
            "= [{}, {}, {}, {}]\n".format(*uf_reds)
        )
    unitline = f"Units:       {units}\n"
    currdate = datetime.date.today().strftime("%d-%b-%Y")
    dateline = f"Date:        {currdate}\n"
    return descline + unceline + unitline + misc + dateline


def _get_numform(mxmn1, excel=False):
    # excel logic is different than text:
    # - it avoids scientific notation since the actual values are
    #   there ... the user can just change the format
    pv = (mxmn1 != 0.0) & np.isfinite(mxmn1)
    if not np.any(pv):
        return "{:13.0f}" if not excel else "#,##0."
    pmx = int(np.floor(np.log10(abs(mxmn1[pv]).max())))
    if excel:
        numform = "#,##0." + "0" * (5 - pmx)
    else:
        pmn = int(np.floor(np.log10(abs(mxmn1[pv]).min())))
        if pmx - pmn < 6 and pmn > -3:
            if pmn < 5:
                numform = f"{{:13.{5 - pmn}f}}"
            else:
                numform = "{:13.0f}"
        else:
            numform = "{:13.6e}"
    return numform


def get_marker_cycle():
    """
    Return an ``itertools.cycle`` of plot markers.

    The list is taken from `matplotlib.markers`. Currently::

        'o',          # circle
        'v',          # triangle_down
        '^',          # triangle_up
        '<',          # triangle_left
        '>',          # triangle_right
        '1',          # tri_down
        '2',          # tri_up
        '3',          # tri_left
        '4',          # tri_right
        '8',          # octagon
        's',          # square
        'p',          # pentagon
        'P',          # plus (filled)
        '*',          # star
        'h',          # hexagon1
        'H',          # hexagon2
        '+',          # plus
        'x',          # x
        'X',          # x (filled)
        'D',          # diamond
        'd',          # thin_diamond

    """
    return itertools.cycle(
        [
            "o",  # circle
            "v",  # triangle_down
            "^",  # triangle_up
            "<",  # triangle_left
            ">",  # triangle_right
            "1",  # tri_down
            "2",  # tri_up
            "3",  # tri_left
            "4",  # tri_right
            "8",  # octagon
            "s",  # square
            "p",  # pentagon
            "P",  # plus (filled)
            "*",  # star
            "h",  # hexagon1
            "H",  # hexagon2
            "+",  # plus
            "x",  # x
            "X",  # x (filled)
            "D",  # diamond
            "d",  # thin_diamond
        ]
    )


def _proc_filterval(filterval, nrows, name="filterval"):
    if filterval is None:
        return None
    filterval = np.atleast_1d(filterval)
    if filterval.ndim > 1:
        raise ValueError(f"`{name}` must be 1-D (is {filterval.ndim}-D)")
    nfilt = len(filterval)
    if nfilt > 1 and nfilt != nrows:
        raise ValueError(
            f"`{name}` has incorrect length:"
            f" expected {nrows} elements, but got {nfilt}"
        )
    return filterval


def PrintCLAInfo(mission, event):
    "Print CLA event info, typically for the log file"
    print(f"Mission:  {mission}")
    print(f"Event:    {event}")


def freq3_augment(freq1, lam, tol=1.0e-5):
    """
    Mimic Nastran's FREQ3 augmentation of a frequency vector.

    Parameters
    ----------
    freq1 : 1d array_like
        Initial frequencies (Hz)
    lam : 1d array_like
        System eigenvalues (rad/sec)^2
    tol : scalar; optional
        Tolerance used for deleting near-duplicate frequencies

    Returns
    -------
    freq : 1d ndarray
        The modified frequency vector

    Notes
    -----
    Only natural frequencies in the range of `freq1` will be added.

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import cla
    >>> freq1 = np.arange(5., 11.)
    >>> sysHz = np.array([3.3, 6.7, 8.9, 9.00001, 12.4])
    >>> lam = (2*np.pi*sysHz)**2
    >>> np.set_printoptions(linewidth=80)
    >>> cla.freq3_augment(freq1, lam)
    array([  5. ,   6. ,   6.7,   7. ,   8. ,   8.9,   9. ,  10. ])
    """
    freq1, lam = np.atleast_1d(freq1, lam)
    sysfreqs = np.sqrt(abs(lam)) / (2 * np.pi)
    pv = np.nonzero(np.logical_and(sysfreqs > freq1[0], sysfreqs < freq1[-1]))[0]
    freq3 = sysfreqs[pv]
    freq = np.sort(np.hstack((freq1, freq3)))
    uniq = locate.find_unique(freq, tol * (freq[-1] - freq[0]))
    return freq[uniq]


def maxmin(response, x):
    """
    Compute max & min of a response matrix.

    Parameters
    ----------
    response : 2d ndarray
        Matrix where each row is a response signal.
    x: 1d ndarray
        X-axis vector (eg, time or frequency);
        ``len(x) = response.shape[1]``

    Returns
    -------
    mxmn : 2d ndarray
        Four column matrix: ``[max, x_of_max, min, x_of_min]``
    """
    r, c = np.shape(response)
    if c != len(x):
        raise ValueError("# of cols in `response` is not compatible with `x`.")
    jx = np.nanargmax(response, axis=1)
    jn = np.nanargmin(response, axis=1)
    ind = np.arange(r)
    mx = response[ind, jx]
    mn = response[ind, jn]
    return SimpleNamespace(
        ext=np.column_stack((mx, mn)), ext_x=np.column_stack((x[jx], x[jn]))
    )


def nan_argmax(v1, v2):
    """
    Find where `v2` is greater than `v1` ignoring NaNs.

    Parameters
    ----------
    v1 : ndarray
        First set of values to compare
    v2 : ndarray
        Second set of values to compare; must be broadcast-compatible
        with `v1`.

    Returns
    -------
    pv : bool ndarray
        Contains True where `v2` is greater than `v1` ignoring NaNs.

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import cla
    >>> v1 = np.array([1.0, np.nan, 2.0, np.nan])
    >>> v2 = np.array([-2.0, 3.0, np.nan, np.nan])
    >>> cla.nan_argmax(v1, v2)
    array([False,  True, False, False], dtype=bool)
    >>> cla.nan_argmin(v1, v2)
    array([ True,  True, False, False], dtype=bool)
    """
    with np.errstate(invalid="ignore"):
        return (v2 > v1) | (np.isnan(v1) & ~np.isnan(v2))


def nan_argmin(v1, v2):
    """
    Find where `v2` is less than `v1` ignoring NaNs.

    Parameters
    ----------
    v1 : ndarray
        First set of values to compare
    v2 : ndarray
        Second set of values to compare; must be broadcast-compatible
        with `v1`.

    Returns
    -------
    pv : bool ndarray
        Contains True where `v2` is less than `v1` ignoring NaNs.

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import cla
    >>> v1 = np.array([1.0, np.nan, 2.0, np.nan])
    >>> v2 = np.array([-2.0, 3.0, np.nan, np.nan])
    >>> cla.nan_argmax(v1, v2)
    array([False,  True, False, False], dtype=bool)
    >>> cla.nan_argmin(v1, v2)
    array([ True,  True, False, False], dtype=bool)
    """
    with np.errstate(invalid="ignore"):
        return (v2 < v1) | (np.isnan(v1) & ~np.isnan(v2))


def nan_absmax(v1, v2):
    """
    Get absolute maximum values between `v1` and `v2` while
    retaining signs and ignoring NaNs.

    Parameters
    ----------
    v1 : ndarray
        First set of values to compare
    v2 : ndarray
        Second set of values to compare; must be broadcast-compatible
        with `v1`.

    Returns
    -------
    amax : ndarray
        The absolute maximum values over `v1` and `v2`. The signs are
        retained and NaNs are ignored as much as possible.
    pv : bool ndarray
        Contains True where ``abs(v2)`` is greater than ``abs(v1)``
        ignoring NaNs.

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import cla
    >>> v1 = np.array([[1, -4], [3, 5]])
    >>> v2 = np.array([[-2, 3], [4, np.nan]])
    >>> cla.nan_absmax(v1, v2)
    (array([[-2, -4],
           [ 4,  5]]), array([[ True, False],
           [ True, False]], dtype=bool))
    """
    amx = v1.copy()
    pv = nan_argmax(abs(v1), abs(v2))
    amx[pv] = v2[pv]
    return amx, pv


def extrema(curext, mm, maxcase, mincase=None, casenum=None):
    """
    Update extrema values in 'curext'

    Parameters
    ----------
    curext : SimpleNamespace
        Has extrema information (members may be None on first call)::

            .ext     = 1 or 2 columns: [max, min]
            .ext_x   = None or 1 or 2 columns: [x_of_max, x_of_min]
            .maxcase = list of strings identifying maximum case
            .mincase = list of strings identifying minimum case

        Also has these members if casenum is an integer::

            .mx      = [case1_max, case2_max, ...]
            .mn      = [case1_min, case2_min, ...]
            .mx_x    = [case1_x_of_max, case2_x_of_max, ...]
            .mn_x    = [case1_x_of_min, case2_x_of_min, ...]

        The `mx_x` and `mn_x` will have ``np.nan`` values if
        `ext_x` is None.
    mm : SimpleNamespace
        Has min/max information for a case (or over cases)::

            .ext     = 1 or 2 columns: [max, min]
            .ext_x   = None or 1 or 2 columns: [x_of_max, x_of_min]

    maxcase : string or list of strings
        String or list of strings identifying the load case(s) for the
        maximum values. This is analogous to `curext.maxcase` but
        pertaining to `mm` (handy if `mm` is from another extrema data
        set).
    mincase : string or list of strings or None; optional
        Analogous to `maxcase` for the minimum values or None. If
        None, it is a copy of the `maxcase` values.
    casenum : integer or None; optional
        If integer, it is case number (starting at 0); `curext` will
        have the `.mx`, `.mn`, `.mx_x`, `.mn_x` members and the
        `casenum` column of each of these will be set to the data from
        `mm`. Zeros will be used for "x" if `mm` is 1 or 2 columns.
        If None, `.mx`, `.mn`, `.mx_x`, `.mn_x` are not updated
        (and need not be present).

    Returns
    -------
    None

    Notes
    -----
    This routine updates the `curext` variable. This routine is
    typically called indirectly via
    :func:`DR_Results.time_data_recovery`,
    :func:`DR_Results.frf_data_recovery`, and
    :func:`DR_Results.psd_data_recovery`.
    """

    def _put_time(curext, mm, j, col_lhs, col_rhs):
        if mm.ext_x is not None:
            if curext.ext_x is None:
                curext.ext_x = copy.copy(mm.ext_x)
            else:
                curext.ext_x[j, col_lhs] = mm.ext_x[j, col_rhs]
        elif curext.ext_x is not None:  # pragma: no cover
            curext.ext_x[j, col_lhs] = np.nan

    r, c = mm.ext.shape
    if c not in [1, 2]:
        raise ValueError(f"mm.ext has {c} cols, but must have 1 or 2.")

    # expand current case information to full size if necessary
    if isinstance(maxcase, str):
        maxcase = r * [maxcase]
    else:
        maxcase = maxcase[:]

    if c == 1:
        if casenum is not None:  # record current results
            curext.mx[:, casenum] = mm.ext[:, 0]
            curext.mn[:, casenum] = mm.ext[:, 0]
            if mm.ext_x is not None:
                curext.mx_x[:, casenum] = mm.ext_x[:, 0]
                curext.mn_x[:, casenum] = mm.ext_x[:, 0]
            else:
                curext.mx_x[:, casenum] = np.nan
                curext.mn_x[:, casenum] = np.nan

        if curext.ext is None:
            curext.ext = mm.ext @ [[1, 1]]
            if mm.ext_x is not None:
                curext.ext_x = mm.ext_x @ [[1, 1]]
            else:
                curext.ext_x = None
            curext.maxcase = maxcase
            curext.mincase = maxcase[:]
            return

        # keep sign but compare based on absolute
        j = nan_argmax(abs(curext.ext), abs(mm.ext)).nonzero()[0]
        if j.size > 0:
            for i in j:
                curext.maxcase[i] = maxcase[i]
            curext.ext[j, 0] = mm.ext[j, 0]
            _put_time(curext, mm, j, 0, 0)

        j = nan_argmin(abs(curext.ext), abs(mm.ext)).nonzero()[0]
        if j.size > 0:
            for i in j:
                curext.mincase[i] = maxcase[i]
            curext.ext[j, 1] = mm.ext[j, 0]
            _put_time(curext, mm, j, 1, 0)
        return

    if mincase is None:
        mincase = maxcase[:]
    elif isinstance(mincase, str):
        mincase = r * [mincase]
    else:
        mincase = mincase[:]

    if casenum is not None:  # record current results
        curext.mx[:, casenum] = mm.ext[:, 0]
        curext.mn[:, casenum] = mm.ext[:, 1]
        if mm.ext_x is not None:
            curext.mx_x[:, casenum] = mm.ext_x[:, 0]
            curext.mn_x[:, casenum] = mm.ext_x[:, 1]
        else:
            curext.mx_x[:, casenum] = np.nan
            curext.mn_x[:, casenum] = np.nan

    if curext.ext is None:
        curext.ext = mm.ext.copy()
        curext.ext_x = copy.copy(mm.ext_x)
        curext.maxcase = maxcase
        curext.mincase = mincase
        return

    j = nan_argmax(curext.ext[:, 0], mm.ext[:, 0]).nonzero()[0]
    if j.size > 0:
        for i in j:
            curext.maxcase[i] = maxcase[i]
        curext.ext[j, 0] = mm.ext[j, 0]
        _put_time(curext, mm, j, 0, 0)

    j = nan_argmin(curext.ext[:, 1], mm.ext[:, 1]).nonzero()[0]
    if j.size > 0:
        for i in j:
            curext.mincase[i] = mincase[i]
        curext.ext[j, 1] = mm.ext[j, 1]
        _put_time(curext, mm, j, 1, 1)


def _calc_covariance_sine_cosine(varx, vary, covar):
    # See, for example:
    # http://www.visiondummy.com/2014/04/
    #             draw-error-ellipse-representing-covariance-matrix/)

    # Have covariance matrix:
    #  A = [varx, covar]
    #      [covar, vary]

    # Need to find the angle from the x-axis to the eigenvector that
    # maximizes the vector sum response (RSS). The maximizing
    # eigenvector corresponds to the largest eigenvalue (can think of
    # this as principal axes too). Could do that with
    # scipy.linalg.eigh:
    #     lam, phi = scipy.linalg.eigh(A)
    #     # since 2nd eigenvalue from `eigh` is biggest:
    #     theta = np.arctan2(phi[1, 1], phi[0, 1])
    #
    # Or, since this is just a 2x2 for each item, we can also solve
    # this by hand ahead of time.

    # allocate sine and cosine arrays:
    n = varx.shape[0]
    s = np.empty(n)
    c = np.empty(n)

    # check for very small covar:
    pv = abs(covar) <= 1e-12 * np.fmax(varx, vary)
    if pv.any():
        # put in pi/2 where vary > varx, 0.0 for others:
        y_bigger = varx[pv] < vary[pv]
        s[pv] = y_bigger  # where vary > varx: sin(pi/2) = 1
        c[pv] = 1 - y_bigger

    pv = ~pv  # where there is a non-zero covariance
    if pv.any():
        varx, vary, covar = varx[pv], vary[pv], covar[pv]
        # Eigenvalues are (from hand calcs):
        #   term = sqrt((vary - varx)**2 + 4*covar**2)
        #   lambda = (varx + vary +- term) / 2
        #
        # Note that both eigenvalues are always real. Since all three
        # values in the lambda expression are positive, the largest
        # eigenvalue is with the positive sign. The eigenvector
        # associated with that is (using the top row of
        # ``(A - lambda_max I) X = 0`` and defining first element as
        # 1):

        #   X = [                1                 ]
        #       [(vary - varx + term) / (2 * covar)]

        # get angle to eigenvector: arctan2(x2, x1) = arctan(x2):
        term = np.sqrt((vary - varx) ** 2 + 4 * covar ** 2)
        theta = np.arctan((vary - varx + term) / (2 * covar))
        s[pv] = np.sin(theta)
        c[pv] = np.cos(theta)

    return s, c


def PSD_consistent_rss(resp, xr, yr, rr, freq, forcepsd, drmres, case, i):
    """
    Compute phase-consistent (time-correlated) root-sum-square (RSS)
    responses in a PSD analysis; each RSS is of two rows.

    In the following, 'x' denotes DOF 1 for the RSS and 'y' denotes
    the perpendicular direction: resp = sqrt(x*x + y*y).

    Parameters
    ----------
    resp : 2d ndarray
        Frequency response curves: DOF x len(freq)
    xr : 1d array_like
        Index vector of the 'x' `resp` rows (input to RSS)
    yr : 1d array_like
        Index vector of the 'y' `resp` rows (input to RSS)
    rr : 1d array_like or None
        If RSS responses are to be stored in the same matrix
        (category) as `resp`, then `rr` is an index vector specifying
        which rows will hold the RSS. For example, `resp` could have
        3 rows with ``xr = 0``, ``yr = 1``, and ``rr = 2``. The 3rd
        row would be the RSS of the first two.

        On the other hand, if the RSS is to be stored alone in its
        category, set `rr` to None. For example, `resp` could have 2
        rows with ``xr = 0``, ``yr = 1``. Then, the category would
        only have 1 data recovery item
    freq : 1d array_like
        Frequency vector (Hz)
    forcepsd : 2d ndarray
        Matrix of force PSDs; nforces x len(freq)
    drmres : input/output SimpleNamespace
        Results for a DRM; eg if drmres = results['ifa']:

        .. code-block:: none

           .rms (r x cases)
           .cases (list of cases)
           .freq (freq x 1)
           .psd (cases x r x freq)
           ._psd (temp dict, psd[case] is freq x r)
           .srs.srs[q] (cases x r x freq)
           .srs.ext[q] (each r x freq)
           .srs.frq (freq,)
           .srs.type (string, either 'srs' or 'eqsine')
           .srs.units (string, eg: 'G, rad/sec^2')
           .ext (r x 2), .maxcase (length r), .mincase ([])

    case : string
        Unique case identifier (like 'MaxQ') for storing the PSD
        results, eg::

            results['ifa']._psd[case]

    i : integer
        Current force index; starts at 0 and goes to nforces-1

    Returns
    -------
    None

    Notes
    -----
    This function is typically called by a "drfunc" specified in a
    call to :func:`DR_Def.add`. An example function that uses this
    routine is provided there.

    The `drmres` input is modified on each call::

         ._psd[case] is updated (size = drm rows x freq)

    On the last call the RSS is computed and stored in ``._psd[case]``
    according to `rr`.

    These members are created to keep track of needed values (but
    deleted on last call)::

         .tmp.varx  = 'x' variance
         .tmp.vary  = 'y' variance
         .tmp.covar = covariance
         .tmp.xresp = list of 'x' response matrices;
                       - list is nforces long
                       - each response matrix is len(xr) x freq
         .tmp.yresp = list of 'y' response matrices

    This routine works by computing the maximum length principal axis
    (eigenvector) of the covariance matrix. The RSS results are then
    calculated along the eigenvector direction. See comments in source
    code of :func:`_calc_covariance_sine_cosine` for more
    details. Also see, for example, reference [#cov]_.

    References
    ----------
    .. [#cov] http://www.visiondummy.com/2014/04/
              draw-error-ellipse-representing-covariance-matrix/
    """
    # drmres is a SimpleNamespace: drmres = results[drm]
    # i is psd force index
    F = forcepsd[i]
    # normal, non-rss data recovery:
    if rr is not None:
        drmres._psd[case] += F * abs(resp) ** 2
    N = forcepsd.shape[0]
    if i == 0:
        drmres.tmp = SimpleNamespace(
            varx=0, vary=0, covar=0, xresp=[0] * N, yresp=[0] * N
        )

    x = resp[xr]
    y = resp[yr]
    tmp = drmres.tmp
    # after area is computed, these will be the variance/covariance:
    tmp.varx += F * abs(x) ** 2
    tmp.vary += F * abs(y) ** 2
    tmp.covar += F * np.real(x * np.conj(y))
    tmp.xresp[i] = x
    tmp.yresp[i] = y

    if i == N - 1:
        varx = np.trapz(tmp.varx, freq)
        vary = np.trapz(tmp.vary, freq)
        covar = np.trapz(tmp.covar, freq)
        s, c = _calc_covariance_sine_cosine(varx, vary, covar)

        # now have all sines/cosines, compute consistent vector sum
        # results:
        rss_resp = 0.0
        for j in range(N):
            respxy = c[:, None] * tmp.xresp[j] + s[:, None] * tmp.yresp[j]
            rss_resp += forcepsd[j] * abs(respxy) ** 2

        if rr is not None:
            drmres._psd[case][rr] = rss_resp
        else:
            drmres._psd[case] = rss_resp

        # delete 'extra' info:
        del drmres.tmp



# -*- coding: utf-8 -*-
"""
Collection of tools used for CLA - coupled loads analysis
"""

from ._utilities import *
from .dr_def import DR_Def
from .dr_event import DR_Event
from .dr_results_plots import mk_plots
from .dr_results import DR_Results, get_drfunc
from ._magpct import magpct
from .rel_disp_dtm import relative_displacement_dtm
from ._rptext1 import rptext1
from ._rpttab1 import rpttab1
from ._rptpct1 import rptpct1
from pyyeti.ytools import save, load



# -*- coding: utf-8 -*-
"""
A pretty printer.
"""
import collections
import numpy as np
import h5py
import pandas as pd


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


class PP(object):
    """
    A simple class for pretty printing data structures.
    """

    def __init__(
        self, var=None, depth=1, tab=4, keylen=40, strlen=80, show_hidden=False
    ):
        """
        Initializer for `PP`.

        Parameters
        ----------
        var : any; optional
            If not None, the variable to pretty print.
        depth : integer; optional
            Maximum number of levels to print.
        tab : integer; optional
            Number of additional indent spaces for each level.
        keylen : integer; optional
            Maximum length for dictionary (and similar) keys.
        strlen : integer; optional
            Maximum length for dictionary (and similar) values.
        show_hidden : bool; optional
            Many objects (classes, class instances, namespaces, etc)
            are printed via looping over the ``.__dict__`` method. For
            those variables, if `show_hidden` is True, show members
            that start with '_'. Default is to not show them.

        Notes
        -----
        A variable can be printed during instantiation or via
        :func:`PP.pp`. For example::

            PP(var, depth=3, tab=8)

        gives the same output as::

            p = PP(depth=3, tab=8)
            p.pp(var)

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> from types import SimpleNamespace
        >>> from pyyeti.pp import PP
        >>> r = np.arange(4, dtype=np.int16)
        >>> s = np.random.randn(4, 4, 4)
        >>> t = np.array(9, dtype=np.uint8)
        >>> d = {'asdf': 4,
        ...      '34': 'value',
        ...      'r': r,
        ...      'dataframe': pd.DataFrame(np.ones((3, 3))),
        ...      'series': pd.Series(np.ones(8)),
        ...      'longer name': {1: 2,
        ...                      2: 3,
        ...                      3: s,
        ...                      4: SimpleNamespace(a=6,
        ...                                         b=[1, 23],
        ...                                         var='string',
        ...                                         t=(s,))
        ...      }
        ... }
        >>> PP(d)       # doctest: +ELLIPSIS
        <class 'dict'>[n=6]
            'asdf'       : 4
            '34'         : 'value'
            'r'          : int16 ndarray 4 elems: (4,) [0 1 2 3]
            'dataframe'  : pandas DataFrame: (3, 3)
            'series'     : pandas Series: (8,)
            'longer name': <class 'dict'>[n=4]
        <BLANKLINE>
        <...>
        >>> PP(d, 5)    # doctest: +ELLIPSIS
        <class 'dict'>[n=6]
            'asdf'       : 4
            '34'         : 'value'
            'r'          : int16 ndarray 4 elems: (4,) [0 1 2 3]
            'dataframe'  : pandas DataFrame: (3, 3)
            'series'     : pandas Series: (8,)
            'longer name': <class 'dict'>[n=4]
                1: 2
                2: 3
                3: float64 ndarray 64 elems: (4, 4, 4)
                4: <class 'types.SimpleNamespace'>[n=4]
                    .a  : 6
                    .b  : [n=2]: [1, 23]
                    .var: 'string'
                    .t  : [n=1]: (float64 ndarray: (4, 4, 4),)
        <BLANKLINE>
        <...>

        To demonstrate the `show_hidden` option:

        >>> class A:
        ...     a = 9
        ...     b = {'variable': [1, 2, 3]}
        >>> PP(A, 2)       # doctest: +ELLIPSIS
        <class 'A'>[n=6]
            .a: 9
            .b: <class 'dict'>[n=1]
                'variable': [n=3]: [1, 2, 3]
        <BLANKLINE>
        <...>
        >>> PP(A, 2, show_hidden=1)        # doctest: +SKIP
        <class 'A'>[n=6]
            .__module__ : 'pyyeti.pp'
            .a          : 9
            .b          : <class 'dict'>[n=1]
                'variable': [n=3]: [1, 2, 3]
            .__dict__   : <attribute '__dict__' of 'A' objects>
            .__weakref__: <attribute '__weakref__' of 'A' objects>
            .__doc__    : None
        <BLANKLINE>
        <...>
        """
        self._tab = tab
        self._depth = depth
        self._keylen = keylen
        self._strlen = strlen
        self._show_hidden = show_hidden
        self._functions = {
            np.ndarray: self._array_string,
            h5py._hl.dataset.Dataset: self._h5data_string,
        }
        if var is not None:
            self.pp(var)

    def _lead_string(self, level):
        return " " * self._tab * level

    def _key_string(self, val, isns):
        if isinstance(val, str):
            if isns:
                s = f".{val}"
            else:
                s = f"'{val}'"
        else:
            s = str(val)
        if len(s) > self._keylen:
            s = f"{s[: self._keylen - 4]} ..."
        return s

    def _lst_tup_string(self, lst, list_level):
        s = [f"[n={len(lst)}]: "]
        be = "[]" if isinstance(lst, list) else "()"
        s.append(be[0])
        if list_level > self._depth:
            s.append(f"...{be[1]}")
            return s
        list_level += 1
        for i, item in enumerate(lst):
            if i > self._strlen // 3:
                # each entry is at least 3 chars: "1, "
                s.append(" ...")
                break
            if isinstance(item, np.ndarray):
                s.extend(self._shortarrhdr(item))
            else:
                s.extend(self._value_string(item, list_level))
            s.append(", ")
        if s[-1] == ", ":
            if len(lst) == 1 and isinstance(lst, tuple):
                s[-1] = ","
            else:
                s = s[:-1]
        s.append(be[1])
        return s

    def _value_string(self, val, list_level):
        if isinstance(val, str):
            s = [f"'{val}'"]
        elif isinstance(val, (list, tuple)):
            s = self._lst_tup_string(val, list_level)
        else:
            s = [f"{val}".replace("\n", " ")]
        s = "".join(s)
        if len(s) > self._strlen:
            n = self._strlen // 2 - 3
            s = f"{s[:n]} ... {s[-n:]}"
        return [s]

    def _shortarrhdr(self, arr):
        return [f"{arr.dtype} ndarray: {arr.shape}"]

    def _getarrhdr(self, arr):
        return [f"{arr.dtype} ndarray {arr.size} elems: {arr.shape}"]

    def _getarrstr(self, arr):
        s = f" {arr}".replace("\n", "")
        if len(s) > 4 * (self._strlen // 5):
            n = self._strlen // 3
            s = f"{s[: n - 3]} <...> {s[-(n + 3) :]}"
        return [s]

    def _array_string(self, arr, level):
        s = self._getarrhdr(arr)
        if arr.size <= 10:
            s.extend(self._getarrstr(arr))
        s.append("\n")
        return s

    def _h5data_string(self, arr, level):
        s = ["H5 "]
        s.extend(self._getarrhdr(arr))
        if arr.size <= 10:
            s.extend(self._getarrstr(arr[...]))
        s.append("\n")
        return s

    def _get_keys(self, dct, showhidden):
        if not showhidden:
            keys = [k for k in dct if k[0] != "_"]
        else:
            keys = list(dct.keys())
        return keys

    def _dict_string(self, dct, level, typename, isns=False, showhidden=True):
        s = [f"{typename}[n={len(dct)}]\n"]
        if level < self._depth:
            keys = self._get_keys(dct, showhidden)
            level += 1
            # get max key length for pretty printing:
            n = 0
            for k in keys:
                n = max(n, len(self._key_string(k, isns)))
            frm = f"{{:<{n}s}}: "
            for k in keys:
                s.append(self._lead_string(level))
                s.append(frm.format(self._key_string(k, isns)))
                s.extend(self._print_var(dct[k], level))
        return s

    def _pandas_string(self, var, level, typename):
        return [f"pandas {var.__class__.__name__}: {var.shape}\n"]

    def _print_var(self, var, level):
        try:
            s = self._functions[type(var)](var, level)
        except KeyError:
            cls = type(var)
            if cls is type:
                typename = f"<class '{var.__name__}'>"
            else:
                typename = str(cls)
            if isinstance(var, collections.Mapping):
                s = self._dict_string(var, level, typename=typename)
            elif isinstance(var, pd.core.base.PandasObject):
                s = self._pandas_string(var, level, typename=typename)
            elif hasattr(var, "__dict__") and not hasattr(var, "keys"):
                s = self._dict_string(
                    var.__dict__,
                    level,
                    typename=typename,
                    isns=True,
                    showhidden=self._show_hidden,
                )
            else:
                s = self._value_string(var, 0)
                s.append("\n")
        return s

    def pp(self, var):
        """
        Pretty print variable `var`. See :class:`PP`.
        """
        s = self._print_var(var, 0)
        self.s = s
        self.output = "".join(s)
        for line in self.output.split("\n"):
            print(repr(line)[1:-1])



# -*- coding: utf-8 -*-
"""
DR_Event: Setup data recovery for a specific event or set of modes
"""
import copy
from types import SimpleNamespace
import warnings
from collections import OrderedDict
import numpy as np
import scipy.linalg as la
from pyyeti.ytools import reorder_dict
from pyyeti.nastran import n2p
from .dr_results import DR_Results
from ._utilities import _merge_uf_reds


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


class DR_Event:
    """
    Setup data recovery for a specific event or set of modes.

    Attributes
    ----------
    Info : :class:`collections.OrderedDict`
        Contains data recovery information for each category. The
        category names are the keys. This is a copy of information in
        one or more `DR_Def` instances created during data recovery
        setup; eg, in a "prepare_4_cla.py" script. See :class:`DR_Def`
        and the example below. The copy is made during the call to
        :func:`DR_Event.add` and new values for the `uf_reds` option
        can be set during that call.
    UF_reds : list
        List of all unique 4-element uncertainty factor tuples. See
        :func:`DR_Def.add` for more information.
    Vars : dict
        Contains the data recovery matrices and possibly other data
        needed for data recovery. This is derived from the
        ``DR_Def['_vars']`` dict and the current system modes. See the
        notes section below for an example showing what is in this
        dict.

    Notes
    -----
    Here are some views of example data via :func:`pyyeti.pp.PP`. This
    would be after, for example::

        DR = cla.DR_Event()
        DR.add(nas, drdefs_for_se100)   # PAF data recovery
        DR.add(nas, drdefs_for_se500)   # SC data recovery
        DR.add(nas, drdefs_for_se0)     # recover "Node4"

    PP(DR, 2):

    .. code-block:: none

        <class 'pyyeti.cla.DR_Event'>[n=3]
            .Info   : <class 'dict'>[n=6]
                'PAF_ifatm': <class 'types.SimpleNamespace'>[n=20]
                'PAF_ifltm': <class 'types.SimpleNamespace'>[n=20]
                'SC_atm'   : <class 'types.SimpleNamespace'>[n=20]
                'SC_dtm'   : <class 'types.SimpleNamespace'>[n=20]
                'SC_ltm'   : <class 'types.SimpleNamespace'>[n=20]
                'Node4'    : <class 'types.SimpleNamespace'>[n=20]
            .UF_reds: [n=2]: [[n=4]: (1, 1, 1.25, 1),
                              [n=4]: (0, 1, 1.25, 1)]
            .Vars   : <class 'dict'>[n=3]
                0  : <class 'dict'>[n=1]
                100: <class 'dict'>[n=3]
                500: <class 'dict'>[n=4]

    PP(DR.Vars, 2):

    .. code-block:: none

        <class 'dict'>[n=3]
            0  : <class 'dict'>[n=1]
                'Tnode4': float64 ndarray 732 elems: (3, 977)
            100: <class 'dict'>[n=3]
                'ifatm' : float64 ndarray 46896 elems: (48, 977)
                'ifltma': float64 ndarray 17586 elems: (18, 977)
                'ifltmd': float64 ndarray 17586 elems: (18, 977)
            500: <class 'dict'>[n=4]
                'scatm' : float64 ndarray 261836 elems: (268, 977)
                'scdtmd': float64 ndarray 84999 elems: (87, 977)
                'scltma': float64 ndarray 142642 elems: (146, 977)
                'scltmd': float64 ndarray 142642 elems: (146, 977)

    In that example, all the ``DR.Vars`` variables except 'Tnode4'
    have been multiplied by the appropriate ULVS matrix in the call to
    :func:`add`. That is, the SE 100 and 500 matrices all came from
    the ``DR_Def['_vars'].drms`` entry and none came from the
    ``DR_Def['_vars'].nondrms`` entry. The SE 0 matrix 'Tnode4' could
    come from either the `.drms` or `.nondrms` entry.
    """

    def __repr__(self):
        cats = ", ".join(f"'{name}'" for name in self.Info)
        return (
            f"{type(self).__name__} ({hex(id(self))}) with "
            f"{len(self.Info)} categories: [{cats}]"
        )

    def __init__(self):
        """
        Initializes the attributes `Info`, `UF_reds`, and `Vars` to
        empty values.

        The attributes are filled by calls to :func:`add`.
        """
        self.Info = OrderedDict()
        self.UF_reds = []
        self.Vars = {}

    def add(self, nas, drdefs, uf_reds=None, method="replace"):
        """
        Add data recovery definitions for an event or set of modes.

        Parameters
        ----------
        nas : dictionary
            This is the nas2cam dictionary:
            ``nas = pyyeti.nastran.op2.rdnas2cam()``. Only used if at
            least one category in `drdefs` is for an upstream
            superelement (se != 0) and has a data recovery matrix. Can
            be anything (like None) if not needed.
        drdefs : :class:`DR_Def` instance or None
            Contains the data recovery definitions for, typically, one
            superelement. See :class:`DR_Def`. If None, this routine
            does nothing.
        uf_reds : 4-element tuple or None; optional
            If not None, this is the uncertainty factors in "reds"
            order: [rigid, elastic, dynamic, static]. The values in
            `uf_reds` either replace or multiply the original values
            (see `method`). Use a value of None for a particular
            uncertainty value to retain the original value
            unmodified. This `uf_reds` option can be useful when
            uncertainty factors are event specific rather than data
            recovery category specific or you need to add in a forcing
            function uncertainty factor.

            For example, to reset the dynamic uncertainty factor to
            1.1 while leaving the other values unchanged::

                uf_reds=(None, None, 1.1, None)
                DR.add(nas, drdefs, uf_reds)

            For another example, to increase the rigid-body and
            elastic uncertainty factors by a factor of 1.05 while
            leaving the other two values unchanged::

                uf_reds=(1.05, 1.05, None, None)
                DR.add(nas, drdefs, uf_reds, method='multiply')

        method : string or function (callable); optional
            Specifies how to update the "uf_reds" settings:

              ============  ========================================
                `method`                 Description
              ============  ========================================
              'replace'     Values in `uf_reds` (that are not None)
                            replace old values.
              'multiply'    Values in `uf_reds` (that are not None)
                            multiply the old values.
               callable     Values are computed by function (or any
                            callable). The function is only called
                            for entries where `uf_reds` is not None.
                            The call is:
                            ``method(old_value, new_value)``. See
                            examples below.
              ============  ========================================

        Notes
        -----
        Typically called once for each superelement where data
        recovery is requested. The attributes `Info`, `UF_reds` and
        `Vars` are all updated on each call to this routine.

        See :class:`DR_Def` for a discussion about how the order of
        data recovery is determined. In summary: :func:`DR_Event.add`
        determines the order of :class:`DR_Def` instances, and
        :func:`DR_Def.add` determines the order of data recovery
        categories within each :class:`DR_Def` instance.
        :func:`DR_Event.set_dr_order` can be used to modify the final
        order.

        Following are some examples of providing a function for the
        `method` parameter. Note that the function is only called when
        the `uf_reds` value is not None.

        These two calls are equivalent::

            DR.add(nas, drdefs, uf_reds, 'replace')
            DR.add(nas, drdefs, uf_reds, lambda old, new: new)

        These are also equivalent::

            DR.add(nas, drdefs, uf_reds, 'multiply')
            DR.add(nas, drdefs, uf_reds, lambda old, new: old*new)

        As a final example, if you wanted to add values onto the
        previous settings instead of multiply, you could do this::

            DR.add(nas, drdefs, uf_reds, lambda old, new: old+new)

        Raises
        ------
        ValueError
            When the there are duplicate category names.
        """
        if drdefs is None:
            return

        for name, drminfo in drdefs.items():
            if name == "_vars":
                continue

            try:
                active = drminfo.active
            except AttributeError:
                active = "yes"

            if active != "yes":
                continue

            if name in self.Info:
                raise ValueError(f'"{name}" data recovery category already defined')

            # variables for how to do data recovery:
            self.Info[name] = copy.copy(drdefs[name])

            # reset uf_reds if input:
            if uf_reds is not None:
                self.Info[name].uf_reds = _merge_uf_reds(
                    self.Info[name].uf_reds, uf_reds, method=method
                )

            # collect all sets of uncertainty factors together for the
            # apply_uf routine later:
            uf_reds_cur = self.Info[name].uf_reds
            if uf_reds_cur not in self.UF_reds:
                self.UF_reds.append(uf_reds_cur)

        se_last = -2
        # apply ULVS to all drms and put in DR:
        for se, dct in drdefs["_vars"].drms.items():
            if not dct:
                continue
            if se not in self.Vars:
                self.Vars[se] = {}
            if se not in (0, se_last):
                ulvs = nas["ulvs"][se]
                uset = nas["uset"][se]
                # Want bset partition from aset. But note that in the
                # .asm, .pch approach to SE's, it is valid in Nastran
                # to just put all b-set & q-set in a generic a-set.
                # If that's the case, find q-set by finding the
                # spoints:
                bset = n2p.mksetpv(uset, "a", "b")  # bool type
                if bset.all():
                    # manually check for q-set in the a-set:
                    aset = np.nonzero(n2p.mksetpv(uset, "p", "a"))[0]
                    dof = uset.index.get_level_values("dof")[aset]
                    qset = dof == 0  # spoints
                    bset = ~qset
                bset = np.nonzero(bset)[0]
                Lb = len(bset)
            se_last = se

            for name, mat in dct.items():
                if name in self.Vars[se]:
                    raise ValueError(f'"{name}" is already in Vars[{se}]')
                if se == 0:
                    self.Vars[se][name] = mat
                elif mat.shape[1] > Lb:
                    self.Vars[se][name] = mat @ ulvs
                else:
                    self.Vars[se][name] = mat @ ulvs[bset]

        # put all nondrms in DR:
        for se, dct in drdefs["_vars"].nondrms.items():
            if se not in self.Vars:
                self.Vars[se] = {}
            for name, mat in dct.items():
                if name in self.Vars[se]:
                    raise ValueError(f'"{name}" is already in Vars[{se}]')
                self.Vars[se][name] = mat

    def set_dr_order(self, cats, where):
        """
        Set a new data recovery order

        Parameters
        ----------
        cats : iterable
            Iterable of category names in the ordered desired for data
            recovery. Not all categories need to be included, just
            those where a new order is desired. For example, if
            "scltm" must be done before "scltm2", then ``cats =
            ('scltm', 'scltm2')`` is sufficient.
        where : string
            Either 'first' or 'last'. Specifies where to put the
            reordered categories in the final order.

        Notes
        -----
        This is a convenience method for the
        :func:`pyyeti.ytools.reorder_dict` routine.  Updates the
        `Info` attribute to reflect the new order. Call this routine
        before calling :func:`prepare_results`.

        See :class:`DR_Def` for a discussion about how the order of
        data recovery is determined. In summary: :func:`DR_Event.add`
        determines the order of :class:`DR_Def` instances, and
        :func:`DR_Def.add` determines the order of data recovery
        categories within each :class:`DR_Def` instance.  This routine
        can be used to modify the final order.

        Raises
        ------
        ValueError
            If an item in `cats` is not currently in `self.Info`
        ValueError
            If `where` is not 'first' or 'last'

        Examples
        --------
        Build a minimal instance of :class:`DR_Event` incorporating
        two :class:`DR_Def` instances so that
        :func:`DR_Event.set_dr_order` can be fully demonstrated. This
        also demonstrates the use of :func:`DR_Def.add` without using
        the :func:`DR_Def.addcat` decorator.

        >>> from pyyeti import cla
        >>>
        >>> drdefs0 = cla.DR_Def()
        >>> for name, nrows in (('atm0', 12),
        ...                     ('ltm0', 30),
        ...                     ('dtm0', 9)):
        ...     drdefs0.add(name=name, labels=nrows, drfunc='no func')
        >>>
        >>> drdefs1 = cla.DR_Def()
        >>> for name, nrows in (('atm1', 12),
        ...                     ('ltm1', 30),
        ...                     ('dtm1', 9)):
        ...     drdefs1.add(name=name, labels=nrows, drfunc='no func')
        >>>
        >>> DR = cla.DR_Event()
        >>> DR.add(None, drdefs1)
        >>> DR.add(None, drdefs0)

        Show that original order is as defined (`drdefs1` is first):

        >>> print(repr(DR))    # doctest: +ELLIPSIS
        DR_Event ... ['atm1', 'ltm1', 'dtm1', 'atm0', 'ltm0', 'dtm0']

        For demonstration, put the two :class:`DR_Def` instances in a
        different order:

        >>> DR = cla.DR_Event()
        >>> DR.add(None, drdefs0)
        >>> DR.add(None, drdefs1)
        >>> print(repr(DR))    # doctest: +ELLIPSIS
        DR_Event ... ['atm0', 'ltm0', 'dtm0', 'atm1', 'ltm1', 'dtm1']

        Ensure that 'ltm1', 'dtm0', 'atm1' are recovered in that
        order, and assume the others are okay in current order:

        Case 1: put 'ltm1', 'dtm0', 'atm1' first:

        >>> DR.set_dr_order(('ltm1', 'dtm0', 'atm1'), where='first')
        >>> print(repr(DR))    # doctest: +ELLIPSIS
        DR_Event ... ['ltm1', 'dtm0', 'atm1', 'atm0', 'ltm0', 'dtm1']

        Case 2: put 'ltm1', 'dtm0', 'atm1' last:

        >>> DR.set_dr_order(('ltm1', 'dtm0', 'atm1'), where='last')
        >>> print(repr(DR))    # doctest: +ELLIPSIS
        DR_Event ... ['atm0', 'ltm0', 'dtm1', 'ltm1', 'dtm0', 'atm1']
        """
        self.Info = reorder_dict(self.Info, cats, where)

    def prepare_results(self, mission, event):
        """
        Returns an instance of the :class:`DR_Results` class.

        Parameters
        ----------
        mission : str
            Identifies the CLA
        event : str
            Name of event

        Returns
        -------
        results : :class:`DR_Results` instance
            Subclass of dict containing categories with results (see
            :class:`DR_Results`).

        Notes
        -----
        Uses the `Info` attribute (see :class:`DR_Event`) and calls
        :func:`DR_Results.init` to build the initial results data
        structure for all data recovery categories.
        """
        results = DR_Results()
        results.init(self.Info, mission, event)
        return results

    def apply_uf(self, sol, m, b, k, nrb, rfmodes):
        """
        Applies the uncertainty factors to the modal ODE solution

        Parameters
        ----------
        sol : SimpleNamespace
            Solution, input only; expected to have::

                .a = modal acceleration time-history matrix
                .v = modal velocity time-history matrix
                .d = modal displacement time-history matrix
                .pg = g-set forces; optional

        m : 1d or 2d ndarray or None
            Modal mass; can be vector or matrix or None (for identity)
        b : 1d or 2d ndarray
            Modal damping; vector or matrix
        k : 1d or 2d ndarray
            Modal stiffness; vector or matrix
        nrb : scalar
            Number of rigid-body modes
        rfmodes : index vector or None
            Specifies where the res-flex modes are; if None, no
            resflex

        Returns
        -------
        solout : dict
            Dictionary of solution namespaces with scaled versions
            of `.a`, `.v`, `.d` and `.pg`. The keys are all the
            "uf_reds" values. Additionally, the displacement member is
            separated into static and dynamic parts: `.d_static`,
            `.d_dynamic`. On output, ``.d = .d_static + .d_dynamic``.
            For example, if one of the "uf_reds" tuples is:
            ``(1, 1, 1.25, 1)``, then these variables will exist::

                solout[(1, 1, 1.25, 1)].a
                solout[(1, 1, 1.25, 1)].v
                solout[(1, 1, 1.25, 1)].d
                solout[(1, 1, 1.25, 1)].d_static
                solout[(1, 1, 1.25, 1)].d_dynamic
                solout[(1, 1, 1.25, 1)].pg (optional)

        Notes
        -----
        Uncertainty factors are applied as follows (rb=rigid-body,
        el=elastic, rf=residual-flexibility)::

           ruf = rb uncertainty factor
           euf = el uncertainty factor
           duf = dynamic uncertainty factor
           suf = static uncertainty factor

           .a_rb and .v_rb - scaled by ruf*suf
           .a_el and .v_el - scaled by euf*duf
           .a_rf and .v_rf - zeroed out

           .d_rb - zeroed out
           .d_el - static part:  scaled by euf*suf
                 - dynamic part: scaled by euf*duf
           .d_rf - scaled by euf*suf

           .pg   - scaled by suf

        Note that d_el is written out as::

              d_el = euf*inv(k_el)*(suf*F_el - duf*(a_el+b_el*v_el))

        where::

              F = m*a + b*v + k*d
        """
        n = k.shape[0]
        use_velo = True

        # genforce = m*a + b*v + k*d
        def _comp_genforce(sol, d):
            if m is None:
                genforce = sol.a.copy()
            elif m.ndim == 1:
                genforce = m[:, None] * sol.a
            else:
                genforce = m @ sol.a

            if b.ndim == 1:
                genforce += b[:, None] * sol.v
            else:
                genforce += b @ sol.v

            if k.ndim == 1:
                genforce += k[:, None] * d
            else:
                genforce += k @ d
            return genforce

        solout = {}
        # This routine will do its work in two loops:
        # - a copy loop
        # - an apply ufs loop
        # The copy loop is separate to be efficient, since copying
        # from 'F' order to 'C' order is slower than 'C' to 'C'
        last = None
        for item in self.UF_reds:
            if last:
                solout[item] = copy.deepcopy(last)
            else:
                # we want C contiguous a, v, d variables for faster uf
                # application
                # - variable 'd' is skipped because it is created
                #   below
                solout[item] = SimpleNamespace()
                for name, value in vars(sol).items():
                    if name == "d":
                        continue
                    try:
                        valcopy = value.copy()
                    except AttributeError:
                        valcopy = copy.deepcopy(value)
                    setattr(solout[item], name, valcopy)
                last = solout[item]
                genforce = _comp_genforce(last, sol.d)

        for item in self.UF_reds:
            ruf, euf, duf, suf = item
            SOL = solout[item]
            SOL.d_static = np.empty_like(SOL.a)
            SOL.d_dynamic = np.empty_like(SOL.a)

            # apply ufs:
            if nrb > 0:
                SOL.a[:nrb] *= ruf * suf
                SOL.v[:nrb] *= ruf * suf
                SOL.d_static[:nrb] = 0.0
                SOL.d_dynamic[:nrb] = 0.0

            if rfmodes is not None:
                SOL.a[rfmodes] = 0.0
                SOL.v[rfmodes] = 0.0

            try:
                SOL.pg *= suf
            except AttributeError:
                pass

            if nrb < n:
                SOL.a[nrb:] *= euf * duf
                SOL.v[nrb:] *= euf * duf

                if m is None:
                    avterm = SOL.a[nrb:].copy()
                elif m.ndim == 1:
                    avterm = m[nrb:, None] * SOL.a[nrb:]
                else:
                    avterm = m[nrb:, nrb:] @ SOL.a[nrb:]

                if not use_velo:  # pragma: no cover
                    msg = (
                        "not including velocity term in mode-"
                        "acceleration formulation for "
                        "displacements."
                    )
                    warnings.warn(msg, RuntimeWarning)
                else:
                    if b.ndim == 1:
                        avterm += b[nrb:, None] * SOL.v[nrb:]
                    else:
                        avterm += b[nrb:, nrb:] @ SOL.v[nrb:]

                # in case there is mass coupling between rfmodes and
                # other modes
                if rfmodes is not None:
                    avterm[rfmodes - nrb] = 0.0

                gf = (euf * suf) * genforce[nrb:]
                if k.ndim == 1:
                    invk = (1 / k[nrb:])[:, None]
                    SOL.d_static[nrb:] = invk * gf
                    SOL.d_dynamic[nrb:, :] = -invk * avterm
                else:
                    lup = la.lu_factor(k[nrb:, nrb:])
                    SOL.d_static[nrb:] = la.lu_solve(lup, gf)
                    SOL.d_dynamic[nrb:, :] = la.lu_solve(lup, -avterm)

                SOL.d = SOL.d_static + SOL.d_dynamic
        return solout

    def frf_apply_uf(self, sol, nrb):
        """
        Applies the uncertainty factors to the frequency response
        functions (FRFs).

        Parameters
        ----------
        sol : SimpleNamespace
            Solution, input only; expected to have::

                .a = modal acceleration FRF matrix
                .v = modal velocity FRF matrix
                .d = modal displacement FRF matrix
                .pg = g-set forces; optional

        nrb : scalar
            Number of rigid-body modes

        Returns
        -------
        solout : dict
            Dictionary of solution namespaces with scaled versions of
            `.a`, `.v`, `.d` and `.pg`. The keys are all the "uf_reds"
            values. For example, if one of the "uf_reds" tuples is:
            ``(1, 1, 1.25, 1)``, then these variables will exist::

                solout[(1, 1, 1.25, 1)].a
                solout[(1, 1, 1.25, 1)].v
                solout[(1, 1, 1.25, 1)].d
                solout[(1, 1, 1.25, 1)].pg (optional)

        Notes
        -----
        Uncertainty factors are applied as follows (rb=rigid-body,
        el=elastic, rf=residual-flexibility)::

           ruf = rb uncertainty factor
           euf = el uncertainty factor
           duf = dynamic uncertainty factor
           suf = static uncertainty factor

           .a_rb, .v_rb, d_rb - scaled by ruf*suf
           .a_el, .v_el, d_el - scaled by euf*duf
           .pg   - scaled by suf
        """
        solout = {}
        for item in self.UF_reds:
            ruf, euf, duf, suf = item
            solout[item] = copy.deepcopy(sol)
            SOL = solout[item]
            SOL.a[:nrb] *= ruf * suf
            SOL.v[:nrb] *= ruf * suf
            SOL.d[:nrb] *= ruf * suf
            SOL.a[nrb:] *= euf * duf
            SOL.v[nrb:] *= euf * duf
            SOL.d[nrb:] *= euf * duf
            if "pg" in SOL.__dict__:
                SOL.pg *= suf
        return solout

    def get_Qs(self):
        """
        Get list of all unique Q's used for SRS in all categories

        Returns
        -------
        Qs : list
            list of all unique Q's used for SRS in all categories
        """
        Qs = set()
        for v in self.Info.values():
            if v.srsQs is not None:
                Qs |= set(v.srsQs)  # or: Qs = Qs.union(v.srsQs)
        return sorted(Qs)



# -*- coding: utf-8 -*-
"""
DR_Def: data recovery definitions
"""
import os
import sys
import copy
import numbers
import inspect
from collections import abc, OrderedDict
from types import SimpleNamespace
import warnings
import copyreg
import numpy as np
import pandas as pd
import xlsxwriter
from pyyeti.nastran import n2p
from ._utilities import _merge_uf_reds, _is_valid_identifier, get_drfunc


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


class DR_Def(OrderedDict):
    """
    Data recovery definitions.

    This class inherits from :class:`collections.OrderedDict` and
    defines how data recovery will be done. The entries are created
    through repeated calls to member function :func:`add` (typically
    from a "prepare_4_cla.py" script). Within a single instance of
    this class, the order that items are added also defines the data
    recovery order (as done in :func:`DR_Results.time_data_recovery`,
    for example). When using multiple instances, the data recovery
    order of the instances is determined by the order they are added
    to an instance of :class:`DR_Event` via :func:`DR_Event.add`.
    :func:`DR_Event.set_dr_order` can be used to modify the final
    order. :func:`pyyeti.ytools.reorder_dict` can also be used for
    that purpose.

    Attributes
    ----------
    defaults : dict or None; optional
        Dictionary with any desired defaults for the parameters
        listed in :func:`add`. If None, it is initialized to an
        empty dictionary.
    ncats : integer
        The number of data recovery categories defined. This is a
        static class variable. This means that if you have multiple
        instances of this class, `ncats` will end up being the total
        number of categories over all instances (and there is nothing
        wrong with that).

    Notes
    -----
    The keys, with the exception of the key '_vars', are the short
    names of each data recovery category. These names are defined by
    the `name` parameter in successive calls to :func:`add`.
    The values for these categories are SimpleNamespaces that fully
    describe the category containing the description, units, row
    labels, shock response spectra requests, etc, and even provides
    (directly or indirectly) the function for performing data recovery
    for the category.

    As noted above, there is one other entry in the dictionary: the
    '_vars' entry. This entry is also a SimpleNamespace with two
    attributes: `drms` and `nondrms`. The `drms` attribute contains
    the data recovery matrices; it is a regular Python dictionary
    indexed by the superelement ID number. The `nondrms` attribute
    contains any other information needed for performing data
    recovery; it is also a regular Python dictionary indexed by the
    superelement ID number.

    For illustration, we'll use :func:`pyyeti.pp.PP` to display
    sections of this dictionary for an example mission:

    PP(drdefs):

    .. code-block:: none

        <class 'cla.DR_Def'>[n=9]
            'SC_atm'    : <class 'types.SimpleNamespace'>[n=20]
            'SC_dtm'    : <class 'types.SimpleNamespace'>[n=20]
            'SC_ifl'    : <class 'types.SimpleNamespace'>[n=20]
            'SC_ltma'   : <class 'types.SimpleNamespace'>[n=20]
            'SC_ltmd'   : <class 'types.SimpleNamespace'>[n=20]
            'SC_cg'     : <class 'types.SimpleNamespace'>[n=20]
            'SC_ifa'    : <class 'types.SimpleNamespace'>[n=20]
            'SC_ifa_0rb': <class 'types.SimpleNamespace'>[n=20]
            '_vars'     : <class 'types.SimpleNamespace'>[n=2]

    PP(drdefs['SC_ifa'], 2):

    .. code-block:: none

        <class 'types.SimpleNamespace'>[n=20]
            .desc      : 'S/C Interface Accelerations'
            .drfile    : '/loads/CLA/Rocket/missions/.../drfuncs.py'
            .drfunc    : 'SC_ifa'
            .filterval : 1e-06
            .histlabels: [n=12]: ['I/F Axial Accel     X sc', ... lv']
            .histpv    : slice(None, 12, None)
            .histunits : 'G, rad/sec^2'
            .ignorepv  : None
            .labels    : [n=12]: ['I/F Axial Accel     X sc', ... lv']
            .misc      : None
            .se        : 500
            .srsQs     : [n=2]: (25, 50)
            .srsconv   : 1.0
            .srsfrq    : float64 ndarray 990 elems: (990,)
            .srslabels : [n=12]: ['$X_{SC}$', '$Y_{SC}$', '$Z_ ...}$']
            .srsopts   : <class 'dict'>[n=2]
                'eqsine': 1
                'ic'    : 'steady'
            .srspv     : slice(None, 12, None)
            .srsunits  : 'G, rad/sec^2'
            .uf_reds   : [n=4]: (1, 1, 1.25, 1)
            .units     : 'G, rad/sec^2'

    PP(drdefs['_vars'], 3):

    .. code-block:: none

        <class 'types.SimpleNamespace'>[n=2]
            .drms   : <class 'dict'>[n=1]
                500: <class 'dict'>[n=7]
                    'SC_ifl_drm': float64 ndarray 1296 ... (12, 108)
                    'SC_cg_drm' : float64 ndarray 1512 ... (14, 108)
                    'SC_ifa_drm': float64 ndarray 1296 ... (12, 108)
                    'scatm'     : float64 ndarray 28944 ... (268, 108)
                    'scdtmd'    : float64 ndarray 9396 ... (87, 108)
                    'scltma'    : float64 ndarray 15768 ... (146, 108)
                    'scltmd'    : float64 ndarray 15768 ... (146, 108)
            .nondrms: <class 'dict'>[n=1]
                500: <class 'dict'>[n=0]
    """

    # static variable:
    ncats = 0

    @staticmethod
    def addcat(func):
        """
        Decorator to ensure :func:`add` is called to add category.

        Notes
        -----
        Example of typical usage::

            @cla.DR_Def.addcat
            def _():
                name = 'SC_atm'
                desc = 'S/C Internal Accelerations'
                units = 'G'
                drms = {name: sc_atm}
                drfunc = f"Vars[se]['{name}'] @ sol.a"
                # ... other variables defined; see :func:`add`.
                drdefs.add(**locals())

        """
        nc = DR_Def.ncats
        func()
        if DR_Def.ncats <= nc:
            msg = (
                "function must call the `DR_Def.add` "
                "method (eg: ``drdefs.add(**locals())``)"
            )
            raise RuntimeError(msg)

    def __init__(self, defaults=None):
        """
        Parameters
        ----------
        defaults : dict or None; optional
            Sets the `defaults` attribute; see :class:`DR_Def`.
        """
        super().__init__()
        self.defaults = {} if defaults is None else defaults
        self["_vars"] = SimpleNamespace(drms={}, nondrms={})
        self._drfilemap = {}

    def __repr__(self):
        cats = ", ".join(f"'{name}'" for name in self)
        return (
            f"{type(self).__name__} ({hex(id(self))}) with "
            f"{len(self) - 1} categories: [{cats}]"
        )

    # add drms and nondrms to self:
    def _add_vars(self, name, drms, nondrms):
        def _add_drms(d1, d2):
            for key in d2:
                if key in d1:
                    # print warning or error out:
                    msg = (
                        f'"{key}" already included for a previously '
                        "defined data recovery matrix"
                    )
                    if d1[key] is d2[key]:
                        warnings.warn(msg, RuntimeWarning)
                    else:
                        raise ValueError("A different " + msg)
                else:
                    d1[key] = d2[key]

        _vars = self["_vars"]
        se = self[name].se
        for curdrms, newdrms in ((_vars.drms, drms), (_vars.nondrms, nondrms)):
            if se not in curdrms:
                curdrms[se] = {}
            _add_drms(curdrms[se], newdrms)

    def _check_for_drfunc(self, filename, funcname):
        def _get_msg():
            s0 = f'When checking for the data recovery function "{funcname}",'
            s1 = (
                "This can be safely ignored if you plan to "
                "implement the data recovery functions later."
            )
            return s0, s1

        try:
            get_drfunc(filename, funcname)
        except (FileNotFoundError, ImportError, SyntaxError) as e:
            s0, s1 = _get_msg()
            msg = (
                f'{s0} import of "{filename}" failed. {s1}\nThe exception was: '
                f"{type(e).__name__}: {e!s}"
            )
            warnings.warn(msg, RuntimeWarning)
        except AttributeError:
            s0, s1 = _get_msg()
            msg = f'{s0} "{funcname}" not found in: {filename}. {s1}'
            warnings.warn(msg, RuntimeWarning)

    @staticmethod
    def _get_pv(pv, name, length):
        if pv is None:
            return pv
        if isinstance(pv, str) and pv == "all":
            return slice(length)
        if isinstance(pv, slice):
            return pv
        if not isinstance(pv, str) and isinstance(pv, (abc.Sequence, numbers.Integral)):
            pv = np.atleast_1d(pv)
        if isinstance(pv, np.ndarray):
            pv = pv.ravel()
            if pv.dtype == bool:
                if len(pv) != length:
                    msg = (
                        f"length of `{name}` ({len(pv)}) does not "
                        f"match length of labels ({length})"
                    )
                    raise ValueError(msg)
                return pv.nonzero()[0]
            elif issubclass(pv.dtype.type, np.integer):
                if pv.max() >= length or pv.min() < 0:
                    msg = (
                        f"values in `{name}` range from [{pv.min()}, {pv.max()}], "
                        f"but should be in range [0, {length}] for this "
                        "category"
                    )
                    raise ValueError(msg)
                return pv
        raise TypeError(f"`{name}` input not understood")

    @staticmethod
    def _get_labels(pv, labels):
        if isinstance(pv, slice):
            return labels[pv]
        return [labels[i] for i in pv]

    def _handle_defaults(self, name):
        """Handle default values and take default actions"""
        # first, the defaults:
        ns = self[name]

        if ns.drfunc is None:
            ns.drfunc = name

        if ns.drfile is None:
            # this is set to defaults only if it is in defaults,
            # otherwise, leave it None
            ns.drfile = self.defaults.get("drfile", None)

        if ns.se is None:
            ns.se = self.defaults.get("se", 0)

        if ns.uf_reds is None:
            ns.uf_reds = self.defaults.get("uf_reds", (1, 1, 1, 1))

        if ns.srsQs is None:
            for k, v in ns.__dict__.items():
                if k.startswith("srs") and v is not None:
                    ns.srsQs = self.defaults
                    break

        if ns.srsfrq is None and ns.srsQs is not None:
            ns.srsfrq = self.defaults

        dct = ns.__dict__
        for key in dct:
            if dct[key] is self.defaults:
                try:
                    dct[key] = self.defaults[key]
                except KeyError:
                    msg = f"{key} set to `defaults` but is not found in `defaults`!"
                    raise ValueError(msg)

        if ns.drfile == ".":
            ns.drfile = None

        # add path to `drfile` if needed:
        if _is_valid_identifier(ns.drfunc):
            try:
                ns.drfile = self._drfilemap[ns.drfile]
            except KeyError:
                calling_file = os.path.realpath(inspect.stack()[2][1])
                if ns.drfile is None:
                    ns.drfile = drfile = calling_file
                else:
                    drfile = ns.drfile
                    callerdir = os.path.dirname(calling_file)
                    ns.drfile = os.path.join(callerdir, drfile)
                self._drfilemap[drfile] = ns.drfile
            self._check_for_drfunc(ns.drfile, ns.drfunc)

        # ensure uf_reds has no None values:
        ns.uf_reds = _merge_uf_reds((1, 1, 1, 1), ns.uf_reds)

        # next, the default actions:
        if ns.desc is None:
            ns.desc = name

        if isinstance(ns.labels, numbers.Integral):
            ns.labels = [f"Row {i + 1:6d}" for i in range(ns.labels)]

        # check filter value:
        try:
            nf = len(ns.filterval)
        except TypeError:
            pass
        else:
            if nf != len(ns.labels):
                raise ValueError(
                    f"length of `filterval` ({nf}) does not match length of"
                    f"labels ({len(ns.labels)})"
                )
            ns.filterval = np.atleast_1d(ns.filterval)

        ns.ignorepv = self._get_pv(ns.ignorepv, "ignorepv", len(ns.labels))

        if ns.srsconv is None and ns.srsQs is not None:
            ns.srsconv = 1.0

        if ns.srsQs is not None:
            ns.srsQs = n2p._ensure_iter(ns.srsQs)
            if ns.srspv is None:
                ns.srspv = slice(len(ns.labels))

        if ns.srspv is not None and ns.srsopts is None:
            ns.srsopts = {}

        # fill in hist-labels/units and srs-labels/units if needed:
        for i in ("hist", "srs"):
            pv = i + "pv"
            dct[pv] = self._get_pv(dct[pv], pv, len(ns.labels))
            if dct[pv] is not None:
                lbl = i + "labels"
                unt = i + "units"
                if dct[lbl] is None:
                    dct[lbl] = self._get_labels(dct[pv], ns.labels)
                if dct[unt] is None:
                    dct[unt] = ns.units

        # ensure that the labels are lists (or None):
        for i in ("", "hist", "srs"):
            lbl = i + "labels"
            if not isinstance(dct[lbl], (list, type(None))):
                dct[lbl] = list(dct[lbl])

    def add(
        self,
        *,
        name,
        labels,
        active="yes",
        drms=None,
        drfunc=None,
        drfile=None,
        se=None,
        desc=None,
        units="Not specified",
        uf_reds=None,
        filterval=1.0e-6,
        histlabels=None,
        histpv=None,
        histunits=None,
        misc=None,
        ignorepv=None,
        nondrms=None,
        srsQs=None,
        srsfrq=None,
        srsconv=None,
        srslabels=None,
        srsopts=None,
        srspv=None,
        srsunits=None,
        **kwargs,
    ):
        """
        Adds a data recovery category.

        All inputs to :func:`add` must be named.

        .. note::

            Any of the inputs can be set to `self.defaults`. In this
            case, the value for the parameter will be taken from the
            `self.defaults` dictionary (which is defined during
            instantiation). A ValueError is raised if `self.defaults`
            does not contain a value for the parameter.

        .. note::

            Using None for some inputs (which is default for many)
            will cause a default action to be taken. These are listed
            below as applicable after "DA:".

        Parameters
        ----------
        name : string
            Short name of data recovery category, eg: 'SC_atm'. This
            typically also defines `drfunc`, the name of the function
            in `drfile` that is called to do data recovery; in that
            case, it must be a valid Python variable name.
        labels : list_like or integer
            List_like of strings describing each row. Can also be an
            integer specifying number of rows being recovered; in this
            case, the list is formed internally as:
            ``['Row 1', 'Row 2', ...]``. This input is used to
            determine number of rows being recovered. If not a list,
            it is converted to a list via :class:`list`.
        active : string; optional
            If 'yes', this category will be included when the
            :func:`DR_Event.add` function is called to add categories
            for the event simulation. Otherwise, this category will be
            ignored during the simulation by default. These tools only
            check for 'yes', but user written tools can make use of
            other settings, perhaps by monkey-patching the
            :class:`DR_Event` class to replace or augment the
            :func:`DR_Event.add` method. As an example, for a category
            that calculates the fairing-to-spacecraft
            loss-of-clearance, you might set `active` to 'no' to
            unconditionally ignore this category but use 'loc' to
            include this category if a fairing is present but ignore
            it if not.
        drms : dict or None; optional
            Dictionary of data recovery matrices for this category;
            keys are matrix names and must match what is used in the
            data recovery function in `drfile`. In the data recovery
            function, these variables are accessed through
            ``Vars[se]``. For example: ``drms = {'scatm': scatm}``
            means that ``Vars[se]['scatm']`` would be used in the
            function (see also `drfunc`). If `se` is greater than 0,
            each matrix, before being stored in ``Vars[se]`` via
            :func:`DR_Event.add`, will be multiplied by the
            appropriate "ULVS" matrix (see
            :func:`pyyeti.nastran.op2.rdnas2cam`) during event
            simulation. If `se` is 0, it is used as is and is likely
            added during event simulation since system modes are often
            needed. Note that when `se` is 0, using `drms` is equivalent
            to using `nondrms`.
        drfunc : string or None; optional
            A string that either defines the data recovery
            calculations directly (a "Type 1" `drfunc`) or, if it is a
            valid Python identifier, it is the name of the data
            recovery function in `drfile` (a "Type 2" `drfunc`). If it
            is Type 2, as a check, this routine will attempt to import
            the function; if not found, a warning is printed.

            Assume that `drfile` is specified properly and the file it
            indicates has::

                def SC_atm(sol, nas, Vars, se):
                    return Vars[se]['atm'] @ sol.a

            (The inputs to the data recovery function are defined
            below in the Notes section.) Further, assume that `name`
            is 'SC_atm'. With these assumptions, we can use either a
            Type 1 or a Type 2 `drfunc`. That is, the following
            settings for `drfunc` are all functionally equivalent:

              =========================  =============================
                     `drfunc`                    Description
              =========================  =============================
              "Vars[se]['atm'] @ sol.a"  Type 1. This defines the
                                         return statement directly
                                         (the function is defined
                                         internally, inside module
                                         :mod:`cla`).
                     'SC_atm'            Type 2. Uses the `SC_atm`
                                         function in `drfile`.
                       None              Type 2. Same as 'SC_atm'.
                                         (Takes the default action ...
                                         see 'DA' below)
              =========================  =============================

            Note that, currently, if the data recovery function needs
            more than just the typical return statement calculation or
            if a PSD-specific data recovery function is needed, then
            `drfunc` must be Type 2.

            DA: set to `name` and `drfile` is used.
        drfile : string or '.' or None; optional
            Only used if needed according to `drfunc` (if using a
            "Type 2" `drfunc`). If needed, `drfile` is the name of
            file that contains the data recovery function named
            `drfunc`. It can optionally also have a PSD-specific data
            recovery function; this must have the same name but with
            "_psd" appended. See notes below for example
            functions. `drfile` is imported during event simulation.
            If not already provided in `drfile`, the full path of
            `drfile` is defined to be relative to the path of the file
            that contains the function that called this routine. If
            input as '.', `drfile` is set to the full name of the file
            that called this routine.

            DA: if `drfile` is set in `self.defaults`, that is used;
            otherwise, it is set to the full name of the file that
            called this routine (as if it was input as '.').
        se : integer or None; optional
            The superelement number.

            DA: get value from `self.defaults` if present; otherwise
            set to 0.
        desc : string or None; optional
            One line description of category; eg:
            ``desc = 'S/C Internal Accelerations'``.

            DA: set to `name`.
        units : string; optional
            Specifies the units.
        uf_reds : 4-element tuple or None; optional
            The uncertainty factors in "reds" order:
            [rigid, elastic, dynamic, static]. Examples::

                (1, 1, 1.25, 1)  - Typical; DUF = 1.25
                (0, 1, 1.25, 1)  - Same, but without rigid-body part
                (1, 1, 0, 1)     - Recover static part only
                (1, 1, 1, 0)     - Recover dynamic part only

            DA: get value from `self.defaults` if present; otherwise
            set to (1, 1, 1, 1). Also, any of the four entries in the
            tuple can be None; these get reset to the corresponding
            entry from the `self.defaults` or, if that's None too, 1.
        filterval : scalar or 1d array_like; optional
            Response values smaller than `filterval` will be skipped
            during comparison to another set of results. If 1d
            array_like, length must be ``len(labels)`` allowing for a
            unique filter value for each row.
        histlabels : list_like or None; optional
            Analogous to `labels` but just for the `histpv` rows.

            DA: derive from `labels` according to `histpv` if
            needed; otherwise, leave it None.
        histpv : 1d array_like or 'all' or slice or None; optional
            Specifies which rows to save the response histories of.
            See note below about inputting partition vectors.
        histunits : string or None; optional
            Units string for the `histpv`.

            DA: set to `units` if `histpv` is not None; otherwise,
            leave it None.
        ignorepv : 1d array_like or 'all' or slice or None; optional
            Typically, this is left as None (default) so that all rows
            are compared. `ignorepv` specifies rows that need to be
            ignored during comparisons against a reference data set.
            If set to 'all', all rows will ignored ... meaning no
            comparisons will be done for this category. See note below
            about inputting partition vectors.
        misc : any object; optional
            Available for storing miscellaneous information for the
            category. It is not used within this module. This option
            is similar to `nondrms` in functionality; the differences
            are that the `misc` input will be stored with the results,
            and that the `nondrms` contents will be put in ``Vars[se]``
            (making those values available in the the data recovery
            function).
        nondrms : dict or None; optional
            With one important exception, this input is used
            identically to `drms`. The exception is that the values in
            `nondrms` are not multiplied by ULVS. Therefore, `nondrms`
            can contain any variables you need for data recovery. An
            alternative option is to include data you need in
            `drfile` with the data recovery functions.
        srsQs : scalar or 1d array_like or None; optional
            Q values for SRS calculation or None.

            DA: set to `self.defaults` if any other `srs*` option is
            not None; otherwise, leave it None.
        srsfrq : 1d array_like or None; optional
            Frequency vector for SRS.

            DA: get value from `self.defaults` if `srsQs` is not None
            (after its default action, if applicable); otherwise,
            leave it None.
        srsconv : scalar or 1d array_like or None; optional
            Conversion factor for the SRS; scalar or vector same
            length as `srspv`.

            DA: it is internally reset to 1.0 if `srsQs` is not None
            (after its default action, if applicable); otherwise,
            leave it None.
        srslabels : list_like or None; optional
            Analogous to `labels` but just for the `srspv` rows.

            DA: derive from `labels` according to `srspv` if needed;
            otherwise, leave it None.
        srsopts : dict or None; optional
            Dictionary of options for SRS; eg:
            ``dict(eqsine=True, ic='steady')``

            DA: set to ``{}`` if `srspv` is not None; otherwise, leave
            it None.
        srspv : 1d array_like or 'all' or slice or None; optional
            Specifies which rows to compute SRS for. See note below
            about inputting partition vectors.

            DA: if `srsQs` is not None (after its default action, if
            applicable), set to ``slice(len(labels))``; otherwise,
            leave it None.
        srsunits : string or None; optional
            Units string for the `srspv`.

            DA: if `srsQs` is not None (after its default action, if
            applicable), set to `units`; otherwise, leave it None.
        **kwargs : dict; optional
            All other inputs are quietly ignored.

        Returns
        -------
        None

        Notes
        -----
        See :class:`DR_Def` for a discussion about how the order of
        data recovery is determined. In summary: :func:`DR_Event.add`
        determines the order of :class:`DR_Def` instances, and
        :func:`DR_Def.add` determines the order of data recovery
        categories within each :class:`DR_Def` instance.
        :func:`DR_Event.set_dr_order` can be used to modify the final
        order. That routine uses the more general
        :func:`.pyyeti.ytools.reorder_dict` to reorder the
        categories.

        **Entering partition vectors.** The `histpv`, `srspv` and
        `ignorepv` inputs are all handled similarly. They can be input
        as either an index or boolean style partition vector. If input
        as 'all', they are reset internally to ``slice(len(labels))``.
        They can also be entered as a slice directly. The stored
        versions are either a slice or an index vector (uses 0-offset
        for standard Python indexing).

        **`drfunc`, `drfile` notes.** If `drfunc` is a valid Python
        identifier (called "Type 2" above), then `drfile` must contain
        the appropriate data recovery function(s) named ``name`` and,
        optionally, ``name_psd``. For a typical data recovery
        category, only one data recovery function would be
        needed. Here are some examples:

        For a typical ATM::

            def SC_atm(sol, nas, Vars, se):
                return Vars[se]['atm'] @ sol.a

        For a typical mode-displacement DTM::

            def SC_dtm(sol, nas, Vars, se):
                return Vars[se]['dtm'] @ sol.d

        For a typical mode-acceleration LTM::

            def SC_ltm(sol, nas, Vars, se):
                return (Vars[se]['ltma'] @ sol.a +
                        Vars[se]['ltmd'] @ sol.d)

        Note that those three examples could also be implemented more
        directly by using the following 3 "Type 1" `drfunc`
        definitions::

            drfunc = "Vars[se]['atm'] @ sol.a"        # ATM
            drfunc = "Vars[se]['dtm'] @ sol.d"        # DTM
            drfunc = \"\"\"(Vars[se]['ltma'] @ sol.a +
                     Vars[se]['ltmd'] @ sol.d)\"\"\"  # Mode-acce LTM

        For a more complicated data recovery category, you'll need to
        use `drfile` and you might also need to include a special PSD
        version. This function has the same name except with
        "_psd" tacked on. For example, to compute a time-consistent
        (or phase-consistent) root-sum-square (RSS), you need to
        provide both functions. Here is a pair of functions that
        recover 3 rows (for `name` "x_y_rss"): acceleration response
        in the 'x' and 'y' directions and the consistent RSS between
        them for the 3rd row. In this example, it is assumed the data
        recovery matrix has 3 rows where the 3rd row could be all
        zeros::

            def x_y_rss(sol, nas, Vars, se):
                resp = Vars[se]['xyrss'] @ sol.a
                xr = 0             # 'x' row(s)
                yr = 1             # 'y' row(s)
                rr = 2             # rss  rows
                resp[rr] = np.sqrt(resp[xr]**2 + resp[yr]**2)
                return resp

            def x_y_rss_psd(sol, nas, Vars, se, freq, forcepsd,
                            drmres, case, i):
                # drmres is a results namespace, eg:
                #   drmres = results['x_y_rss']
                # i is psd force number
                resp = Vars[se]['xyrss'] @ sol.a
                xr = 0             # 'x' row(s)
                yr = 1             # 'y' row(s)
                rr = 2             # rss  rows
                cla.PSD_consistent_rss(
                    resp, xr, yr, rr, freq, forcepsd,
                    drmres, case, i)

        Ultimately, the PSD recovery function must provide the final
        PSD solution in ``drmres[case]`` after the final force is
        analyzed (when ``i == forcepsd.shape[0]-1``). In the example
        above, :func:`PSD_consistent_rss` takes care of that job.

        Data recovery function inputs:

            =====  ===================================================
            Input  Description
            =====  ===================================================
            sol    ODE modal solution namespace with uncertainty
                   factors applied. Typically has at least .a, .v and
                   .d members (modal accelerations, velocities and
                   displacements). See
                   :func:`pyyeti.ode.SolveUnc.tsolve`.
            nas    The nas2cam dict:
                   ``nas = pyyeti.nastran.op2.rdnas2cam()``
            DR     Defines data recovery for an event simulation (and
                   is created in the simulation script via
                   ``DR = cla.DR_Event()``). It is an event specific
                   version of all combined :class:`DR_Def` objects
                   with all ULVS matrices applied.
            se     Superelement number. Used as key into `DR`.
            etc    See :func:`PSD_consistent_rss` for description of
                   `freq`, `forcepsd`, etc.
            =====  ===================================================
        """
        # get dictionary of inputs and trim out specially handled
        # entries:
        frame = inspect.currentframe()
        # inspect.getargvalues is deprecated
        # args, _, _, values = inspect.getargvalues(frame)
        co = frame.f_code
        nargs = co.co_argcount + co.co_kwonlyargcount
        args = co.co_varnames[:nargs]
        values = frame.f_locals
        dr_def_args = {
            i: values[i]
            for i in args
            if i not in ("self", "name", "drms", "nondrms", "kwargs")
        }
        if name in self:
            raise ValueError(f'data recovery for "{name}" already defined')

        if drms is None:
            drms = {}

        if nondrms is None:
            nondrms = {}

        # check for overlapping keys in drms and nondrms:
        overlapping_names = set(nondrms) & set(drms)
        if overlapping_names:
            raise ValueError(
                f"`drms` and `nondrms` have overlapping names: {overlapping_names}"
            )

        # define self[name] entry:
        self[name] = SimpleNamespace(**dr_def_args)

        # hand defaults and default actions:
        self._handle_defaults(name)

        # add drms and nondrms to self:
        self._add_vars(name, drms, nondrms)

        # increment static variable ncats for error checking:
        DR_Def.ncats += 1

    def amend(self, *, name, overwrite_drms=False, **kwargs):
        """
        Amend a category

        Parameters
        ----------
        name : string
            Name of category to amend, eg: 'SC_atm'. Must already
            exist (if it doesn't, just use :func:`add` to create it as
            desired).
        overwrite_drms : bool; optional
            Allow replacement of original `drms` and `nondrms`. This
            can be dangerous, since multiplie categories can use the
            same `drms` and `nondrms`; use with caution.
        **kwargs : dict; optional
            Any inputs to amend. Any unrecognized entries are quietly
            ignored.

        Returns
        -------
        None

        Notes
        -----
        This routine collects all original settings for category
        `name`, replaces any of them by whatever is specified in
        ``**kwargs``, and calls :func:`add` to add the category to a
        temporary DR_Def instance. If there are no errors, the old
        category is replaced with the new one. Original data recovery
        order is maintained in case that is important.
        """
        ns = self[name]

        # up-front check for drms/nondrms:
        for drm in ("drms", "nondrms"):
            if drm in kwargs and not overwrite_drms:
                raise ValueError(
                    f"to overwrite the '{drm}' entry, set `overwrite_drms` to True"
                )

        # work with a temporary instance:
        temp_drdefs = DR_Def(self.defaults)

        # copy old dictionary:
        settings = ns.__dict__.copy()
        settings.update(kwargs)

        # add updated category to temporary DR_Def:
        temp_drdefs.add(name=name, **settings)

        # replace old category with updated version:
        self[name] = temp_drdefs[name]

        # check for drms/nondrms:
        se = self[name].se
        for drm in ("drms", "nondrms"):
            if drm in kwargs:
                new_dct = getattr(temp_drdefs["_vars"], drm)[se]
                if len(new_dct) > 0:
                    old_dct = getattr(self["_vars"], drm)[se]
                    old_dct.update(new_dct)

    def copycat(self, categories, name_addon, **kwargs):
        """
        Copy a category with optional modifications.

        Parameters
        ----------
        categories : string or list
            Name or names of category(s) to copy.
        name_addon : string or list
            If string, it will be appended to all category names to
            make the new names. If list, contains the new names with
            no regard to old names.
        **kwargs : misc
            Any options to modify. See :func:`add` for complete list.
            For `uf_reds`, set any of the four values to None to keep
            the original value.

        Returns
        -------
        None
            Adds a data recovery category.

        Notes
        -----
        The data recovery categories are copies of the originals with
        the `name` changed (according to `name_addon`) and new values
        set according to `**kwargs`.

        One very common usage is to add a zero-rigid-body version of a
        category; for example::

            drdefs.copycat(['SC_ifa', 'SC_atm'], '_0rb',
                             uf_reds=(0, None, None, None))

        would add 'SC_ifa_0rb' and 'SC_atm_0rb' copy categories
        without the rigid-body component. (Note that, as a
        convenience, :func:`add_0rb` exists for this specific task.)

        For another example, recover the 'SC_cg' (cg load factors)
        in "static" and "dynamic" pieces::

            drdefs.copycat('SC_cg', '_static',
                             uf_reds=(None, None, 0, None))
            drdefs.copycat('SC_cg', '_dynamic',
                             uf_reds=(None, None, None, 0))

        As a final example to show the alternate use of `name_addon`,
        here is an equivalent call for the static example::

            drdefs.copycat('SC_cg', ['SC_cg_static'],
                             uf_reds=(None, None, 0, None))

        Raises
        ------
        ValueError
            When the new category name already exists.
        """
        if isinstance(categories, str):
            categories = [categories]

        for name in categories:
            if name not in self:
                raise ValueError(f"{name} not found")

        for key in kwargs:
            if key not in self[categories[0]].__dict__:
                raise ValueError(f'{key} not found in `self["{categories[0]}"]`')

        for i, name in enumerate(categories):
            if isinstance(name_addon, str):
                new_name = name + name_addon
            else:
                new_name = name_addon[i]
            if new_name in self:
                raise ValueError(f'"{new_name}" category already defined')
            self[new_name] = copy.copy(self[name])
            cat = self[new_name]
            for key, value in kwargs.items():
                if key == "uf_reds":
                    cat.__dict__[key] = _merge_uf_reds(self[name].uf_reds, value)
                else:
                    cat.__dict__[key] = value

    def add_0rb(self, *args):
        """
        Add zero-rigid-body versions of selected categories.

        Parameters
        ----------
        *args : strings
            Category names for which to make '_0rb' versions.

        Notes
        -----
        This is a convenience function that uses :func:`copycat` to do
        the work::

            copycat(args, '_0rb', uf_reds=(0, None, None, None),
                    desc=desc+' w/o RB')

        where `desc` is the current value. See :func:`copycat` for
        more information.

        For example::

            drdefs.add_0rb('SC_ifa', 'SC_atm')

        would add 'SC_ifa_0rb' and 'SC_atm_0rb' categories with the
        first element in `uf_reds` set to 0.
        """
        for arg in args:
            if arg not in self:
                raise ValueError(f"{arg} not found")
            desc = self[arg].desc + " w/o RB"
            self.copycat(arg, "_0rb", uf_reds=(0, None, None, None), desc=desc)

    def excel_summary(self, excel_file="dr_summary.xlsx"):
        """
        Make excel file with summary of data recovery information.

        Parameters
        ----------
        excel_file : string or None; optional
            Name of excel file to create; if None, no file will be
            created.

        Returns
        -------
        drinfo : pandas DataFrame
            Summary table of data recovery information. The index
            values are those set via :func:`add` (name, desc, etc).
            The columns are the categories (eg: 'SC_atm', 'SC_ltm',
            etc).
        """
        cats = sorted([i for i in self if not i.startswith("_")])
        if not cats:
            raise RuntimeError("add data recovery categories first")
        vals = sorted(self[cats[0]].__dict__)
        df = pd.DataFrame(index=vals, columns=cats)

        def _issame(old, new):
            if new is old:
                return True
            if type(new) != type(old):
                return False
            if isinstance(new, np.ndarray):
                if new.shape != old.shape:
                    return False
                return (new == old).all()
            return new == old

        fill_char = "-"
        # fill in DataFrame, use `fill_char` for "same as previous"
        for i, cat in enumerate(cats):
            for val in vals:
                new = self[cat].__dict__[val]
                s = None
                if i > 0:
                    old = self[cats[i - 1]].__dict__[val]
                    if _issame(old, new):
                        self[cat].__dict__[val] = old
                        s = fill_char
                if s is None:
                    s = str(new)
                    if len(s) > 80:
                        s = s[:35] + " ... " + s[-35:]
                    if not isinstance(new, str):
                        try:
                            slen = len(new)
                        except TypeError:
                            pass
                        else:
                            s = f"{slen}: {s}"
                df[cat].loc[val] = s

        if excel_file is not None:
            with xlsxwriter.Workbook(excel_file) as workbook:
                hform = workbook.add_format(
                    {"bold": True, "align": "center", "valign": "vcenter", "border": 1}
                )
                tform = workbook.add_format({"border": 1, "text_wrap": True})
                worksheet = workbook.add_worksheet("DR_Def")
                worksheet.set_column("A:A", 10)
                worksheet.set_column(1, len(cats), 25)
                # worksheet.set_default_row(20)
                # write header:
                for i, cat in enumerate(df.columns):
                    worksheet.write(0, i + 1, cat, hform)
                # write labels:
                for i, lbl in enumerate(df.index):
                    worksheet.write(i + 1, 0, lbl, hform)
                # write table:
                for i, cat in enumerate(df.columns):
                    for j, lbl in enumerate(df.index):
                        worksheet.write(j + 1, i + 1, df[cat].loc[lbl], tform)
                # write notes at bottom:
                bold = workbook.add_format({"bold": True})
                worksheet.write(df.shape[0] + 2, 1, "Notes:", bold)
                tab = "    "
                msg = fill_char + " = same as previous category"
                worksheet.write(df.shape[0] + 3, 1, tab + msg)
                msg = (
                    "The partition vector variables (*pv) "
                    "use 0-offset (or are slices)"
                )
                worksheet.write(df.shape[0] + 4, 1, tab + msg)
                # freeze row 0 and column 0:
                worksheet.freeze_panes(1, 1)
        return df

    @staticmethod
    def merge(first, *args):
        """
        Merge DR_Def instances together to make a new instance

        Parameters
        ----------
        first : DR_Def instance
            The first DR_Def instance to merge with other instances
        *args : any number of DR_Def instances
            The other DR_Def instances to merge with `first`

        Returns
        -------
        DR_Def instance
            The merger of all input DR_Def instances

        Notes
        -----
        The final order of categories is as input. That is, the
        categories of `first` are first and in the order they were
        defined. The others are treated similarly. Note that the order
        can be changed via :func:`pyyeti.ytools.reorder_dict`.
        :func:`DR_Event.set_dr_order` can also be used to modify the
        final data recovery order.

        The ``+`` operator can also be used to merge DR_Def instances.
        That is, this::

            drdefs_new = DR_Def.merge(drdefs1, drdefs2, drdefs3)

        is equivalent to this::

            drdefs_new = drdefs1 + drdefs2 + drdefs3

        except calling :func:`DR_Def.merge` directly is more efficient
        if there are more than two instances being merged.

        Raises
        ------
        ValueError
            When the there are duplicate category names.

        Examples
        --------
        For demonstration, create two data recovery categories using
        two DR_Def instances instead of just one, and them merge them
        together:

        >>> from pyyeti import cla
        >>>
        >>> drdefs1 = cla.DR_Def()
        >>> @cla.DR_Def.addcat
        ... def _():
        ...     se = 101
        ...     name = 'atm'
        ...     desc = 'description'
        ...     labels = 12
        ...     drms = {'atm': 1}
        ...     nondrms = {'radius': 10}
        ...     drfunc = f"Vars[se]['{name}'] @ sol.a"
        ...     drdefs1.add(**locals())
        >>>
        >>> drdefs2 = cla.DR_Def()
        >>> @cla.DR_Def.addcat
        ... def _():
        ...     se = 102
        ...     name = 'dtm'
        ...     desc = 'description'
        ...     labels = 6
        ...     drms = {'dtm': 2}
        ...     nondrms = {'station': 20}
        ...     drfunc = f"Vars[se]['{name}'] @ sol.d"
        ...     drdefs2.add(**locals())
        >>>
        >>> drdefs = cla.DR_Def.merge(drdefs1, drdefs2)
        >>> drdefs    # doctest: +ELLIPSIS
        DR_Def (...) with 2 categories: ['_vars', 'atm', 'dtm']

        You can also use the ``+`` operator:

        >>> drdefs = drdefs1 + drdefs2
        >>> drdefs    # doctest: +ELLIPSIS
        DR_Def (...) with 2 categories: ['_vars', 'atm', 'dtm']
        """
        # use a set to check for duplicate category names:
        cats = set(first)
        cats.remove("_vars")
        duplicate_cats = set()
        for arg in args:
            newcats = set(arg)
            newcats.remove("_vars")
            duplicate_cats |= cats.intersection(newcats)
            cats |= newcats

        # were there any duplicate categories?:
        if duplicate_cats:
            raise ValueError(
                f"there were duplicate categories:\n\t{sorted(duplicate_cats)!r}"
            )

        # category names are unique, need to check drms & nondrms:
        data = {}
        for attr_name in ("drms", "nondrms"):
            duplicate_drms = {}
            drms = getattr(first["_vars"], attr_name).copy()
            for arg in args:
                new_drms = getattr(arg["_vars"], attr_name)
                for se, drm_dict in new_drms.items():
                    if se in drms:
                        # check for duplicate drm names:
                        current = set(drms[se])
                        new = set(drm_dict)
                        if se not in duplicate_drms:
                            duplicate_drms[se] = set()
                        duplicate_drms[se] |= current.intersection(new)
                        drms[se].update(drm_dict)
                    else:
                        drms[se] = drm_dict.copy()

            # were there any duplicate drm names?:
            dup_names = "".join(
                f"{se}: {sorted(dups)}\n" for se, dups in duplicate_drms.items() if dups
            )

            if dup_names:
                raise ValueError(
                    f'there were duplicate "{attr_name}" names. By SE:\n{dup_names}'
                )

            data[attr_name] = drms

        # if here, no duplicates, so merge:
        drdefs = DR_Def()
        drdefs.update(first)
        for arg in args:
            drdefs.update(arg)
        drdefs["_vars"] = SimpleNamespace(drms=data["drms"], nondrms=data["nondrms"])
        return drdefs

    def __add__(self, other):
        """
        Syntactic sugar for the :func:`DR_Def.merge` method.
        """
        return DR_Def.merge(self, other)


# setup pickling for a little bit of future-proofing:
def unpickle_drdefs(kwargs):
    # pickle_version is not used yet
    kwargs.pop("__pickle_version", 0)

    # Cannot use old drdefaults. For example, the drdefs may have been
    # merged from multiple different sets, all with their own
    # "drdefaults". So, get it but don't use it
    drdefaults = kwargs.pop("__defaults", None)

    # Currently, "srsQs" and "srsfrq" are set to "defaults" if None
    # (which is handy when defining categories originally, but less
    # handy for unpickling already-processed categories) ... then, if
    # any other "srs*" value is set (like "srsconv" used to be), the
    # "add" will fail because they're not in the "defaults". This is
    # avoided here simply by setting these two values to None in a new
    # "defaults". (This feels messy ... I'd like a better way to do
    # handle this ....)
    new_drdefs = DR_Def({"srsQs": None, "srsfrq": None})
    for k, v in kwargs.items():
        if k.startswith("__"):
            continue
        if k.startswith("_"):
            new_drdefs[k] = v
        else:
            new_drdefs.add(name=k, **v.__dict__)

    if drdefaults:
        new_drdefs.defaults = drdefaults

    return new_drdefs


def pickle_drdefs(drdefs):
    odct = OrderedDict(drdefs)
    odct["__defaults"] = drdefs.defaults
    odct["__pickle_version"] = 1
    return unpickle_drdefs, (odct,)


copyreg.pickle(DR_Def, pickle_drdefs)



# -*- coding: utf-8 -*-
"""
Power spectral density tools.
"""

from warnings import warn
import numpy as np
import scipy.signal as signal
from scipy.interpolate import interp1d
from pyyeti import dsp


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


def _set_frange(frange, low, high):
    s = frange[0] if frange[0] > 0.0 else low
    e = frange[-1] if frange[-1] <= high else high
    return s, e


def get_freq_oct(n, frange=(1.0, 10000.0), exact=False, trim="outside", anchor=None):
    r"""
    Get frequency vector on an octave scale.

    Parameters
    ----------
    n : scalar
        Specify octave band:  1 for full octave, 3 for 1/3 octave, 6
        for 1/6, etc.
    frange : 1d array_like; optional
        Specifies bounds for the frequencies according to input
        `trim`. Only the first and last elements are used. If the
        first element <= 0.0, 1.0 is used instead. See also the `trim`
        input.
    trim : string; optional
        Specify how to trim frequency vector to `frange`:

        =========   ================================================
         `trim`     Description
        =========   ================================================
        'inside'    All frequencies inside `frange`:
                        ``F_lower[0] >= frange[0]``;
                        ``F_upper[-1] <= frange[-1]``
        'center'    Center frequencies inside `frange`:
                        ``F[0] >= frange[0]``;
                        ``F[-1] <= frange[-1]``
        'outside'   First band includes ``frange[0]`` and last band
                    includes ``frange[-1]``
        'band'      Same as 'outside'
        =========   ================================================

    exact : bool; optional
        If False, return an approximate octave scale so that it hits
        the power of 10s, achored at 1 Hz by default (see
        `anchor`). If True, return an exact octave scale, anchored at
        1000 Hz by default.
    anchor : scalar or None; optional
        If scalar, it specifies the anchor. If None, the anchor used
        is specified under `exact` above (1 or 1000).

    Returns
    -------
    F : 1d ndarray
        Contains the band center frequencies on an octave scale.
    F_lower : 1d ndarray
        Same size as `F`, band lower frequencies.
    F_upper : 1d ndarray
        Same size as `F`, band upper frequencies.

    Notes
    -----
    If `exact` is False, the center, lower, and upper frequencies are
    computed from (where :math:`i` and :math:`j` are integers
    according to `frange` and `trim`):

    .. math::
        \begin{aligned}
        F &= anchor \cdot 10^{3 [i, i+1, i+2, ..., j] / (10 n)} \\
        F_{lower} &= F/10^{3/(20 n)} \\
        F_{upper} &= F \cdot 10^{3/(20 n)}
        \end{aligned}

    If `exact` is True:

    .. math::
        \begin{aligned}
        F &= anchor \cdot 2^{[i, i+1, i+2, ..., j] / n} \\
        F_{lower} &= F/2^{1/(2 n)} \\
        F_{upper} &= F \cdot 2^{1/(2 n)}
        \end{aligned}

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import psd
    >>> np.set_printoptions(precision=4, linewidth=75)
    >>> np.array(psd.get_freq_oct(3, [505, 900]))
    array([[  501.1872,   630.9573,   794.3282,  1000.    ],
           [  446.6836,   562.3413,   707.9458,   891.2509],
           [  562.3413,   707.9458,   891.2509,  1122.0185]])
    >>> np.array(psd.get_freq_oct(3, [505, 900], trim='center'))
    array([[ 630.9573,  794.3282],
           [ 562.3413,  707.9458],
           [ 707.9458,  891.2509]])
    >>> np.array(psd.get_freq_oct(3, [505, 900], exact=True))
    array([[  500.    ,   629.9605,   793.7005,  1000.    ],
           [  445.4494,   561.231 ,   707.1068,   890.8987],
           [  561.231 ,   707.1068,   890.8987,  1122.462 ]])
    >>> psd.get_freq_oct(6, [.8, 2.6])[0]
    array([ 0.7943,  0.8913,  1.    ,  1.122 ,  1.2589,  1.4125,  1.5849,
            1.7783,  1.9953,  2.2387,  2.5119])
    >>> psd.get_freq_oct(6, [.8, 2.6], anchor=2)[0]
    array([ 0.7962,  0.8934,  1.0024,  1.1247,  1.2619,  1.4159,  1.5887,
            1.7825,  2.    ,  2.244 ,  2.5179])
    >>> psd.get_freq_oct(6, [.8, 2.6], exact=True)[0]
    array([ 0.7751,  0.87  ,  0.9766,  1.0962,  1.2304,  1.3811,  1.5502,
            1.74  ,  1.9531,  2.1923,  2.4608])
    >>> psd.get_freq_oct(6, [.8, 2.6], exact=True, anchor=2)[0]
    array([ 0.7937,  0.8909,  1.    ,  1.1225,  1.2599,  1.4142,  1.5874,
            1.7818,  2.    ,  2.2449,  2.5198])
    """
    s = frange[0] if frange[0] > 0.0 else 1.0
    e = frange[-1]
    if exact:
        if not anchor:
            anchor = 1000.0
        var1 = np.floor(np.log2(s / anchor) * n)
        var2 = np.log2(e / anchor) * n + 1
        bands = np.arange(var1, var2)
        F = anchor * 2 ** (bands / n)
        factor = 2 ** (1 / (2 * n))
    else:
        if not anchor:
            anchor = 1.0
        var1 = np.floor(np.log10(s / anchor) * 10 * n / 3)
        var2 = np.log10(e / anchor) * 10 * n / 3 + 1
        bands = np.arange(var1, var2)
        F = anchor * 10 ** (3 * bands / (10 * n))
        factor = 10 ** (3 / (20 * n))
    FL, FU = F / factor, F * factor
    if trim in ("outside", "band"):
        Nmax = np.max(np.nonzero(FL <= e)[0]) + 1
        Nmin = np.min(np.nonzero(FU >= s)[0])
    elif trim == "center":
        Nmax = np.max(np.nonzero(F <= e)[0]) + 1
        Nmin = np.min(np.nonzero(F >= s)[0])
    elif trim == "inside":
        Nmax = np.max(np.nonzero(FU <= e)[0]) + 1
        Nmin = np.min(np.nonzero(FL >= s)[0])
    else:
        raise ValueError("input `trim` has invalid value")
    F = F[Nmin:Nmax]
    FL = FL[Nmin:Nmax]
    FU = FU[Nmin:Nmax]
    return F, FL, FU


def proc_psd_spec(spec):
    """
    Process PSD specification for other routines

    Parameters
    ----------
    spec : 2d ndarray or 2-element tuple/list
        If ndarray, its columns are ``[Freq, PSD1, PSD2, ... PSDn]``.
        Otherwise, it must be a 2-element tuple or list, eg:
        ``(Freq, PSD)`` where PSD is: ``[PSD1, PSD2, ... PSDn]``. In
        the second usage, PSD can be 1d; in the first usage, PSD is
        always considered 2d.

    Returns
    -------
    freq : 1d ndarray
        The frequency vector in `spec`.
    PSD : 1d or 2d ndarray
        The PSD matrix or vector as noted above. Will be 1d only if
        the second usage of `spec` was used and `PSD` is 1d. Shape is
        either ``(len(freq), n)`` or ``(len(freq),)``.
    npsds : integer
        Number of PSDs in `PSD`.

    Notes
    -----
    Any NaNs in the specification frequency are deleted (along with
    the corresponding PSD values).
    """
    if isinstance(spec, np.ndarray):
        if spec.ndim != 2 or spec.shape[1] <= 1:
            raise ValueError(
                "incorrectly sized ndarray for "
                "`spec` input (must be 2d with more"
                " than 1 column)"
            )
        Freq = spec[:, 0]
        PSD = spec[:, 1:]
        npsds = PSD.shape[1]
    else:
        if len(spec) != 2:
            msg = (
                "incorrectly sized `spec` input: for tuple/"
                f"list input, must have length 2, not {len(spec)}"
            )
            raise ValueError(msg)
        Freq, PSD = np.atleast_1d(*spec)
        if len(Freq) != PSD.shape[0]:
            raise ValueError("Freq and PSD inputs in `spec` are incompatibly sized")
        if PSD.ndim > 2:
            raise ValueError("the PSD input in `spec` has more than 2 dimensions.")
        npsds = 1 if PSD.ndim == 1 else PSD.shape[1]
    # check for nans in Freq:
    pv = np.isnan(Freq)
    if pv.any():
        pv = ~pv
        Freq = Freq[pv]
        PSD = PSD[pv]
    return Freq, PSD, npsds


def area(spec):
    r"""
    Compute the area under a PSD random specification.

    Parameters
    ----------
    spec : 2d ndarray or 2-element tuple/list
        If ndarray, its columns are ``[Freq, PSD1, PSD2, ... PSDn]``.
        Otherwise, it must be a 2-element tuple or list, eg:
        ``(Freq, PSD)`` where PSD is: ``[PSD1, PSD2, ... PSDn]``. In
        the second usage, PSD can be 1d; in the first usage, PSD is
        always considered 2d. The frequency vector must be
        monotonically increasing.

    Returns
    -------
    area : 1d array
        Area(s) under the PSD specification.

    Notes
    -----
    This routine assumes a constant db/octave slope for all segments.
    With this assumption, it computes the area for each segment with
    the following formula (derivation below). The segment goes from
    (f1, p1) to (f2, p2)::

         s = log(p2/p1)/log(f2/f1)
         if s == -1:
             area_segment = p1*f1*log(f2/f1)
         else:
             area_segment = (f2*p2 - f1*p1)/(s + 1)

    .. warning::
        This routine is only for specifications with all constant
        db/octave segments. Do not use for general freq vs. psd
        curves, such as output from analysis; use something like
        :func:`numpy.trapz` instead.

    The following derives the equations for computing the area under
    the curve. Each segment is assumed to have a constant db/octave
    slope. We'll start by computing the slope (:math:`m`) of each
    segment. To get the slope, we'll need the number of dbs :math:`d`
    and the number of octaves :math:`n`. First, the number of dbs:

    .. math::
        d = 10 \log_{10} (p_2 / p_1)

    The number of octaves (:math:`n`) is defined by :math:`f_2 = 2^n
    f_1`. Solving for :math:`n` gives:

    .. math::
        n = \frac{\log_{10} (f_2 / f_1)}{\log_{10} (2)}

    Therefore, the slope :math:`m = d / n` for the segment is:

    .. math::
        m = 10 \log_{10} (2) \frac{\log_{10} (p_2 / p_1)}
            {\log_{10} (f_2 / f_1)}

    To simplify the equations, we'll define the variable :math:`s` as:

    .. math::
        s = \frac{m}{10 \log_{10} (2)} = \frac{\log_{10} (p_2 / p_1)}
            {\log_{10} (f_2 / f_1)}

    Solving for :math:`p_2` gives:

    .. math::
        p_2 = p_1 ( f_2 / f_1 )^s

    Since the segment has a constant db/octave slope, that equation
    can be generalized for any frequency value :math:`x` from
    :math:`f_1` to :math:`f_2`:

    .. math::
        p(x) = p_1 ( x / f_1 )^s

    The area under the segment is simply the integral of that
    equation:

    .. math::
        area_{segment} = \int_{f_1}^{f_2} p_1 ( x / f_1 )^s dx

    If :math:`s = -1`:

    .. math::
        area_{segment} = p_1 f_1 \ln ( f_2 / f_1 )

    Otherwise, if :math:`s \neq -1`:

    .. math::
        \begin{aligned}
        area_{segment} &= \frac{p_1 ( f_2^{s+1} - f_1^{s+1} )}
        {f_1^s (s+1)} \\
        &= ( p_2 f_2 - p_1 f_1 ) / (s + 1)
        \end{aligned}

    Examples
    --------
    Compute a 3-sigma peak value given a test spec:

    >>> import numpy as np
    >>> from pyyeti import psd
    >>> spec = np.array([[20, .0053],
    ...                  [150, .04],
    ...                  [600, .04],
    ...                  [2000, .0036]])
    >>> 3*np.sqrt(psd.area(spec))   # doctest: +ELLIPSIS
    array([ 18.43...])

    For comparison, use a brute-force technique:

    >>> f = np.arange(20, 2000.1, 0.1)
    >>> p = psd.interp(spec, f, linear=False)
    >>> 3*np.sqrt(np.trapz(p, f, axis=0))   # doctest: +ELLIPSIS
    array([ 18.43...])
    """
    Freq, PSD, _ = proc_psd_spec(spec)
    if PSD.ndim == 1:
        PSD = PSD[:, None]
    _area = np.zeros(PSD.shape[1])

    # accumulate the areas of all segments for each curve:
    for i in range(Freq.size - 1):
        f1 = Freq[i]
        f2 = Freq[i + 1]
        for j in range(PSD.shape[1]):
            p1 = PSD[i, j]
            p2 = PSD[i + 1, j]

            s = np.log(p2 / p1) / np.log(f2 / f1)
            if abs(s + 1.0) < 1e-5:
                # happens when p2/p1 = f1/f2
                #   slope = -10*log10(2) db/octave
                intarea = p1 * f1 * np.log(f2 / f1)
            else:
                intarea = (f2 * p2 - f1 * p1) / (s + 1.0)
            _area[j] += intarea
    return _area


def interp(spec, freq, linear=False):
    """
    Interpolate values on a PSD specification (or analysis curve).

    Parameters
    ----------
    spec : 2d ndarray or 2-element tuple/list
        If ndarray, its columns are ``[Freq, PSD1, PSD2, ... PSDn]``.
        Otherwise, it must be a 2-element tuple or list, eg:
        ``(Freq, PSD)`` where PSD is: ``[PSD1, PSD2, ... PSDn]``. In
        the second usage, PSD can be 1d; in the first usage, PSD is
        always considered 2d. The frequency vector must be
        monotonically increasing.
    freq : 1d array
        Vector of frequencies to interpolate the specification to.
    linear : bool
        If True, use linear interpolation to expand `spec` to the
        frequencies in `freq`. If False, `spec` is expanded via
        interpolation in log space. In other words:

        ================   ==========================================
        Use:               When:
        ================   ==========================================
        ``linear=False``   `spec` is an actual PSD test specification
                           -- that is, it uses constant db/octave
                           slopes
        ``linear=True``    `spec` doesn't use constant db/octave
                           slopes (eg, an analysis curve)
        ================   ==========================================

    Returns
    -------
    psdfull : 1d or 2d array
        Matrix of the interpolated PSD values. If 2d, has "n" columns
        (which will be one less than `spec` in the first usage above
        because the frequency column is not included). Will be 1d if
        PSD was 1d on input.

    Notes
    -----
    For PSD data that is defined on a center-band frequency scale, use
    :func:`rescale` instead.

    Zeros are used to fill in PSD values for frequencies outside the
    specification(s).

    Any NaNs in the specification frequency are deleted (along with
    the corresponding PSD values) before interpolation. This is done
    by the routine :func:`proc_psd_spec`.

    See also
    --------
    :func:`rescale`

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import psd
    >>> spec = np.array([[20, .0053],
    ...                  [150, .04],
    ...                  [600, .04],
    ...                  [2000, .0036]])
    >>> freq = [100, 200, 600, 1200]
    >>> np.set_printoptions(precision=3)
    >>> psd.interp(spec, freq).ravel()
    array([ 0.027,  0.04 ,  0.04 ,  0.01 ])
    >>> psd.interp(spec, freq, linear=True).ravel()
    array([ 0.027,  0.04 ,  0.04 ,  0.024])

    Using the second form of the spec input:

    >>> spec = ([   20, 150, 600,  2000],
    ...         [.0053, .04, .04, .0036])
    >>> psd.interp(spec, freq, linear=True)
    array([ 0.027,  0.04 ,  0.04 ,  0.024])
    """
    Freq, PSD, npsds = proc_psd_spec(spec)
    #    spec = np.atleast_2d(spec)
    freq = np.atleast_1d(freq)
    if linear:
        ifunc = interp1d(
            Freq, PSD, axis=0, bounds_error=False, fill_value=0, assume_sorted=True
        )
        psdfull = ifunc(freq)
    else:
        ifunc = interp1d(
            np.log(Freq),
            np.log(PSD),
            axis=0,
            bounds_error=False,
            fill_value=0,
            assume_sorted=True,
        )
        psdfull = ifunc(np.log(freq))
        pv = (freq >= Freq[0]) & (freq <= Freq[-1])
        psdfull[pv] = np.exp(psdfull[pv])
    return psdfull


def rescale(P, F, n_oct=3, freq=None, extendends=True, frange=None):
    """
    Convert PSD from one frequency scale to another.

    Parameters
    ----------
    P : array_like
        Vector or matrix; PSD(s) to convert. Works columnwise if
        matrix.
    F : array_like
        Vector; center frequencies for `P`. If steps are not linear,
        logarithmic spacing is assumed.
    n_oct : scalar; optional
        Specifies the desired octave scale. 1 means full octave, 3
        means 1/3rd octave, 6 means 1/6th octave, etc. The routine
        :func:`get_freq_oct` is used to calculate the frequency vector
        with the default options. To change options, call
        :func:`get_freq_oct` directly and provide that input via
        `freq`.
    freq : array_like or None; optional
        Alternative to `n_oct` and takes precedence. Specifies
        desired output frequencies directly. If steps are not linear,
        logarithmic spacing is assumed for computing the band
        boundaries.
    extendends : bool; optional
        If True and the first and/or last frequency band extends
        beyond the original data, the area of the new band is adjusted
        up by the ratio of the new bandwidth over the portion that is
        covered by the original data. This will cause the
        corresponding PSD value to be higher than it would otherwise
        be, meaning the overall mean-square value will also be a
        little higher.
    frange : 1d array_like or None; optional
        This option can be used to limit the frequency range of the
        output frequencies. If None, this option is ignored for cases
        where `freq` is used, but is set internally to ``(25.0,
        np.inf))`` for cases where `n_oct` is used. Only the first and
        last elements of `frange` are used. Note that the output
        frequencies will be trimmed further if needed by the first and
        last values of `F`.

        .. note::
            When the `n_oct` option is used, if the first value of
            `frange` is <= 0.0, it is internally reset to 1.0.

    Returns
    -------
    Pout : ndarray
        Vector or matrix; converted PSD(s). Rows correspond to the
        new frequency scale; columns correspond with `P`.
    Fctr : ndarray
        Vector; center frequencies of output. Equal to `freq` if that
        was input.
    msv : scalar or 1d ndarray
        Mean square value estimate for each PSD.
    ms : ndarray
        Vector or matrix; converted mean square values directly
        (instead of density). For constant frequency step df,
        ``ms = df*Pout``. Same size as `Pout`.

    Notes
    -----
    The input PSD is assumed to be on a center-band frequency
    scale. For PSD specifications that use constant dB/octave slopes,
    use :func:`interp` instead.

    This algorithm works by interpolating on cummulative area such
    that original contributions to total mean-square per band is
    preserved.

    .. note::

        Note that if the area of the first and/or last band is
        extended (see `extendends` above), the overall mean-square
        value will be higher than the original.

    See :func:`get_freq_oct` for more information on how the octave
    scales are calculated.

    See also
    --------
    :func:`interp`

    Examples
    --------
    Compute a PSD with :func:`scipy.signal.welch` and then rescale it
    to 3rd, 6th, and 12th octave scales starting at 1.0 Hz. Compare
    mean square values from all 4 PSDs.

    .. plot::
        :context: close-figs

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import scipy.signal as signal
        >>> from pyyeti import psd
        >>>
        >>> frange = (1.0, np.inf)
        >>> g = np.random.randn(10000)
        >>> sr = 400
        >>> f, p = signal.welch(g, sr, nperseg=sr)
        >>> p3, f3, msv3, ms3 = psd.rescale(p, f, frange=frange)
        >>> p6, f6, msv6, ms6 = psd.rescale(
        ...     p, f, n_oct=6, frange=frange)
        >>> p12, f12, msv12, ms12 = psd.rescale(
        ...     p, f, n_oct=12, frange=frange)
        >>>
        >>> fig = plt.figure('Example')
        >>> fig.clf()
        >>> line = plt.semilogx(f, p, label='Linear')
        >>> line = plt.semilogx(f3, p3, label='1/3 Octave')
        >>> line = plt.semilogx(f6, p6, label='1/6 Octave')
        >>> line = plt.semilogx(f12, p12, label='1/12 Octave')
        >>> _ = plt.legend(ncol=2, loc='best')
        >>> _ = plt.xlim([1, 200])
        >>> msv1 = np.sum(p[1:]*(f[1]-f[0]))
        >>> abs(msv1/msv3 - 1) < .12
        True
        >>> abs(msv1/msv6 - 1) < .06
        True
        >>> abs(msv1/msv12 - 1) < .03
        True

    Demonstrate ``extendends=False``. Ends will be less than input
    because there is a region of zero area averaged in:

    >>> in_freq = np.arange(0, 10.1, .25)
    >>> out_freq = np.arange(0, 10.1, 5)
    >>> in_p = np.ones_like(in_freq)
    >>> p, f, ms, mvs = psd.rescale(in_p, in_freq, freq=out_freq)
    >>> p
    array([ 1.,  1.,  1.])
    >>> p, f, ms, mvs = psd.rescale(in_p, in_freq, freq=out_freq,
    ...                             extendends=False)
    >>> p
    array([ 0.525,  1.   ,  0.525])

    The 0.525 value is from:
    ``area/width = 1*(2.5-(-0.125))/5 = 0.525``
    """
    F = np.atleast_1d(F)
    P = np.atleast_1d(P)

    def _get_fl_fu(fcenter):
        Df = np.diff(fcenter)
        # if np.all(Df == Df[0]):
        if (abs(Df / Df[0] - 1.0) < 1e-12).all():
            # linear scale
            Df = Df[0]
            FL = fcenter - Df / 2
            FU = fcenter + Df / 2
        else:
            # output is not linear, assume log
            mid = np.sqrt(fcenter[:-1] * fcenter[1:])
            fact1 = mid[0] / fcenter[1]
            fact2 = fcenter[-1] / mid[-1]
            FL = np.hstack((fact1 * fcenter[0], mid))
            FU = np.hstack((mid, fact2 * fcenter[-1]))
        return FL, FU

    if freq is None:
        if frange is None:
            frange = (25.0, np.inf)
        frange = _set_frange(frange, 1.0, F[-1])
        Wctr, FL, FU = get_freq_oct(n_oct, exact=True, frange=frange)
    else:
        freq = np.atleast_1d(freq)
        if frange is not None:
            freq = freq[(freq >= frange[0]) & (freq <= frange[-1])]
        FL, FU = _get_fl_fu(freq)
        Nmax = np.max(np.nonzero(FL <= F[-1])[0]) + 1
        Nmin = np.min(np.nonzero(FU >= F[0])[0])
        Wctr = freq[Nmin:Nmax]
        FL = FL[Nmin:Nmax]
        FU = FU[Nmin:Nmax]

    oned = False
    if P.ndim == 1 or (P.ndim == 2 and P.shape[0] == 1):
        oned = True
        P = P.reshape(-1, 1)
        cols = 1
    else:
        cols = P.shape[1]

    # calculate cumulative area:
    Df = np.diff(F)
    if np.all(Df == Df[0]):
        # input uses linear frequency scale
        FLin = F - Df[0] / 2
        FUin = F + Df[0] / 2
    else:
        # not linear, assume log
        FLin, FUin = _get_fl_fu(F)

    Df = (FUin - FLin).reshape(-1, 1)
    ca = np.vstack((np.zeros((1, cols)), np.cumsum(Df * P, axis=0)))
    Fa = np.hstack((FLin[0], FUin))

    if extendends:
        fl = FL[0]
        fu = FU[-1]
        if FL[0] < FLin[0]:
            FL[0] = FLin[0]
        if FU[-1] > FUin[-1]:
            FU[-1] = FUin[-1]

    cal = np.zeros((len(FL), cols))
    cau = np.zeros((len(FL), cols))
    for i in range(cols):
        # with np.interp, interpolating cumulative area beyond end
        # points will take the end value -- that's perfect here: 0's
        # on the front, total area on the back
        cal[:, i] = np.interp(FL, Fa, ca[:, i])
        cau[:, i] = np.interp(FU, Fa, ca[:, i])

    # Compute new values
    ms = cau - cal
    psdoct = ms * (1 / (FU - FL).reshape(-1, 1))
    if extendends:
        FL[0] = fl
        FU[-1] = fu
        ms = psdoct * (FU - FL).reshape(-1, 1)
    msv = np.sum(ms, axis=0)
    if oned:
        psdoct = psdoct.ravel()
        ms = ms.ravel()
        msv = msv[0]
    return psdoct, Wctr, msv, ms


def spl(
    x,
    sr,
    nperseg=None,
    overlap=0.5,
    window="hanning",
    timeslice=1.0,
    tsoverlap=0.5,
    fs=3,
    pref=2.9e-9,
    frange=(25.0, np.inf),
):
    r"""
    Sound pressure level estimation using PSD.

    Parameters
    ----------
    x : 1d array like
        Vector of pressure values.
    sr : scalar
        Sample rate.
    nperseg : int, optional
        Length of each segment for the FFT. Defaults to
        ``int(sr / 5)`` for 5 Hz frequency step in PSD. Note:
        frequency step in Hz = ``sr/nperseg``.
    overlap : scalar; optional
        Amount of overlap in windows, eg 0.5 would be 50% overlap.
    window : str or tuple or array like; optional
        Passed to :func:`scipy.signal.welch`; see that routine for
        more information.
    timeslice : scalar or string-integer
        If scalar, it is the length in seconds for each slice. If
        string, it contains the integer number of points for each
        slice. For example, if `sr` is 1000 samples/second,
        ``timeslice=0.75`` is equivalent to ``timeslice="750"``.
    tsoverlap : scalar in [0, 1) or string-integer
        If scalar, is the fraction of each time-slice to overlap. If
        string, it contains the integer number of points to
        overlap. For example, if `sr` is 1000 samples/second,
        ``tsoverlap=0.5`` and ``tsoverlap="500"`` each specify 50%
        overlap.
    fs : integer; optional
        Specifies output frequency scale. Zero means linear scale,
        anything else is passed to :func:`rescale`. Example:

        ===  ======================================================
          0  linear scale as computed by :func:`scipy.signal.welch`
          1  full octave scale
          3  3rd octave scale
          6  6th octave scale
        ===  ======================================================

    pref : scalar; optional
        Reference pressure. 2.9e-9 psi is considered to be the
        threshhold for human hearing. In Pascals, that value is 2e-5.
    frange : 1d array_like; optional
        Specifies bounds for the frequencies. Only the first and last
        elements are used. If the first value is < 0.0, it is reset
        to 0.0. If the last value is > ``sr/2``, it is reset to
        ``sr/2``. Note that for octave scales, :func:`rescale` is used
        which enforces a minimum of 1.0 Hz.

    Returns
    -------
    f : ndarray
        The output frequency vector (Hz).
    spl : ndarray
        The sound pressure level vector in dB.
    oaspl : scalar
        The overall sound pressure level.

    Notes
    -----
    This routine ultimately calls :func:`scipy.signal.welch` (via
    :func:`psdmod`) to calculate the PSD. It then converts that to
    mean-square values per band (for linear frequency steps, this is
    just ``PSD * delta_freq``). Loosely, the math is:

    .. math::
        \begin{aligned}
        V &= \frac{PSD \cdot \Delta f}{P_{ref}^2} \\
        SPL &= 10 \; \log_{10} ( V ) \\
        OASPL &= 10 \; \log_{10} \left ( \sum V \right )
        \end{aligned}

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import psd
    >>> x = np.random.randn(100000)
    >>> sr = 4000
    >>> f, spl, oaspl = psd.spl(x, sr, sr, timeslice=len(x)/sr)
    >>> # oaspl should be around 170.75 (since variance = 1):
    >>> shouldbe = 10*np.log10(1/(2.9e-9)**2)
    >>> abs(oaspl/shouldbe - 1) < .01
    True
    """
    if nperseg is None:
        nperseg = int(sr / 5)
    # compute psd
    F, P = psdmod(
        x,
        sr,
        nperseg=nperseg,
        window=window,
        timeslice=timeslice,
        tsoverlap=tsoverlap,
        noverlap=int(overlap * nperseg),
    )
    s, e = _set_frange(frange, 0.0, sr / 2)
    if fs != 0:
        _, F, _, P = rescale(P, F, n_oct=fs, frange=(s, e))
    else:
        P = P * F[1]
        pv = (F >= s) & (F <= e)
        F = F[pv]
        P = P[pv]
    v = P / pref ** 2
    return F, 10 * np.log10(v), 10 * np.log10(np.sum(v))


def psd2time(
    spec, ppc, fstart, fstop, df, winends=None, gettime=False, expand_method="interp"
):
    """
    Generate a 'random' time domain signal given a PSD specification.

    Parameters
    ----------
    spec : 2d ndarray or 2-element tuple/list
        If ndarray, its columns are ``[Freq, PSD1, PSD2, ... PSDn]``.
        Otherwise, it must be a 2-element tuple or list, eg:
        ``(Freq, PSD)`` where PSD is: ``[PSD1, PSD2, ... PSDn]``. In
        the second usage, PSD can be 1d; in the first usage, PSD is
        always considered 2d.
    ppc : scalar
        Points per cycle at highest (`fstop`) frequency; if < 2, it is
        internally reset to 2.
    fstart : scalar
        Starting frequency in Hz
    fstop : scalar
        Stopping frequency in Hz
    df : scalar
        Frequency step to be represented in the time signal. If
        routine gives poor results, try refining `df`. If `df` is
        greater than `fstart`, it is reset internally to `fstart`.
    winends : None or dictionary; optional
        If None, :func:`pyyeti.dsp.windowends` is not
        called. Otherwise, `winends` must be a dictionary of arguments
        that will be passed to :func:`pyyeti.dsp.windowends` (not
        including `signal`).
    gettime : bool; optional
        If True, a time vector is output.
    expand_method : str; optional
        Either 'interp' or 'rescale', referring to which function in
        this module will be used to expand input `spec` to all needed
        frequencies. Use 'interp' if `spec` is a specification with
        constant dB/octave slopes. Use 'rescale' if `spec` provides
        the PSD levels on a center-band scale. See :func:`interp` and
        :func:`rescale` for more information.

    Returns
    -------
    sig : 1d ndarray
        The time domain signal with properties set to match input PSD
        spectrum.
    sr : scalar
        The sample rate of the signal.
    time : 1d ndarray; optional
        Time vector for the signal starting at zero with step of
        ``1/sr``: ``time = np.arange(len(sig))/sr``

    Raises
    ------
    ValueError
        On invalid setting for `expand_method`.

    See also
    --------
    :func:`interp`, :func:`rescale`, :func:`pyyeti.dsp.windowends`

    Examples
    --------
    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> from pyyeti import psd
        >>> spec = np.array([[20,  .0768],
        ...                  [50,  .48],
        ...                  [100, .48]])
        >>> sig, sr = psd.psd2time(spec, ppc=10, fstart=35,
        ...                        fstop=70, df=.01,
        ...                        winends=dict(portion=.01))
        >>> sr  # the sample rate should be 70*10 = 700
        700.0

        Compare the resulting psd to the spec from 37 to 68:

        >>> import matplotlib.pyplot as plt
        >>> import scipy.signal as signal
        >>> f, p = signal.welch(sig, sr, nperseg=sr)
        >>> pv = np.logical_and(f >= 37, f <= 68)
        >>> fi = f[pv]
        >>> psdi = p[pv]
        >>> speci = psd.interp(spec, fi).ravel()
        >>> abs(speci - psdi).max() < .05
        True
        >>> abs(np.trapz(psdi, fi) - np.trapz(speci, fi)) < .25
        True
        >>> fig = plt.figure('Example')
        >>> fig.clf()
        >>> a = plt.subplot(211)
        >>> line = plt.plot(np.arange(len(sig))/sr, sig)
        >>> plt.grid(True)
        >>> a = plt.subplot(212)
        >>> line = plt.loglog(spec[:, 0], spec[:, 1], label='spec')
        >>> line = plt.loglog(f, p, label='PSD of time signal')
        >>> leg = plt.legend(loc='best')
        >>> x = plt.xlim(20, 100)
        >>> y = plt.ylim(.05, 1)
        >>> plt.grid(True)
        >>> xticks = np.arange(20, 105, 10)
        >>> x = plt.xticks(xticks, xticks)
        >>> yticks = (.05, .1, .2, .5, 1)
        >>> y = plt.yticks(yticks, yticks)
        >>> v = plt.axvline(35, color='black', linestyle='--')
        >>> v = plt.axvline(70, color='black', linestyle='--')
    """
    if df > fstart:
        df = fstart
    if ppc < 2:
        ppc = 2
    # compute parameters
    # 1 cycle of lowest frequency defines length of signal:
    T = 1 / df  # number of seconds for lowest frequency cycle
    N = int(np.ceil(fstop * ppc * T))  # total number of pts
    df = fstop * ppc / N
    # adjust df to make sure fstart is an even multiple of df
    df = fstart / np.floor(fstart / df)
    sr = N * df  # sr = N/T = N/(1/df)
    odd = N & 1

    # define constants
    freq = np.arange(fstart, fstop + df / 2, df)

    # generate amp(f) vector
    if expand_method == "interp":
        speclevel = interp(spec, freq).ravel()
    elif expand_method == "rescale":
        _freq, _psd, npsds = proc_psd_spec(spec)
        speclevel, *_ = rescale(_psd.ravel(), _freq, freq=freq)
    else:
        raise ValueError(
            '`expand_method` must be either "interp" or "rescale", '
            f"not {expand_method!r}"
        )

    amp = np.sqrt(2 * speclevel * df)

    m = (N + 1) // 2 if odd else N // 2 + 1

    # build up amp to include zero frequency to fstart and from fstop
    # to fhighest:
    ntop = int(np.floor((fstart - df / 2) / df) + 1)
    nbot = m - ntop - len(amp)
    if nbot > 0:
        amp = np.hstack((np.zeros(ntop), amp, np.zeros(nbot)))
    else:
        amp = np.hstack((np.zeros(ntop), amp))

    # generate F(t)
    phi = np.random.rand(m) * np.pi * 2  # random phase angle

    # force these terms to be pure cosine
    phi[0] = 0.0
    if not odd:
        phi[m - 1] = 0

    # coefficients:
    a = amp * np.cos(phi)
    b = -amp * np.sin(phi)

    # form matrix ready for ifft:
    if odd:
        a2 = a[1:m] / 2
        b2 = b[1:m] / 2
        r = N * np.hstack((a[0], a2, a2[::-1]))  # real part
        i = N * np.hstack((0, b2, -b2[::-1]))  # imag part
    else:
        a2 = a[1 : m - 1] / 2
        b2 = b[1 : m - 1] / 2
        r = N * np.hstack((a[0], a2, a[m - 1], a2[::-1]))  # real part
        i = N * np.hstack((0, b2, 0, -b2[::-1]))  # imag part

    F_time = np.fft.ifft(r + 1j * i)
    mxi = abs(F_time.imag).max()
    mxr = abs(F_time.real).max()
    if mxi > 1e-7 * mxr:  # pragma: no cover
        # bug in the code if this ever happens
        warn(
            f"method failed accuracy test; (max imag)/(max real) = {mxi / mxr}",
            RuntimeWarning,
        )

    F_time = F_time.real
    if winends is not None:
        F_time = dsp.windowends(F_time, **winends)
    if gettime:
        return F_time, sr, np.arange(N) / sr
    return F_time, sr


def psdmod(sig, sr, nperseg=None, timeslice=1.0, tsoverlap=0.5, getmap=False, **kwargs):
    """
    Modified method for PSD estimation via FFT.

    Parameters
    ----------
    sig : 1d array_like
        Time series of measurement values.
    sr : scalar
        Sample rate.
    nperseg : int, optional
        Length of each segment for the FFT. Defaults to
        ``int(sr / 5)`` for 5 Hz frequency step in PSD. Note:
        frequency step in Hz = ``sr/nperseg``.
    timeslice : scalar or string-integer
        If scalar, it is the length in seconds for each slice. If
        string, it contains the integer number of points for each
        slice. For example, if `sr` is 1000 samples/second,
        ``timeslice=0.75`` is equivalent to ``timeslice="750"``.
    tsoverlap : scalar in [0, 1) or string-integer
        If scalar, is the fraction of each time-slice to overlap. If
        string, it contains the integer number of points to
        overlap. For example, if `sr` is 1000 samples/second,
        ``tsoverlap=0.5`` and ``tsoverlap="500"`` each specify 50%
        overlap.
    getmap : bool, optional
        If True, get the PSD map output (the `Pmap` and `t` variables
        described below).
    *kwargs : optional
        Named arguments to pass to :func:`scipy.signal.welch`.

    Returns
    -------
    f : 1d ndarray
        Array of sample frequencies.
    Pxx : 1d ndarray
        Power spectral density or power spectrum of `sig`.
    Pmap : 2d ndarray; optional
        The PSD map; each column is an output of
        :func:`scipy.signal.welch`. Rows correspond to frequency `f`
        and columns correspond to time `t`. Only output if `getmap` is
        True.
    t : 1d ndarray; optional
        The time vector for the columns of `Pmap`. Only output if
        `getmap` is True.

    Notes
    -----
    This routine calls :func:`pyyeti.dsp.waterfall` for handling the
    timeslices and preparing the output and :func:`scipy.signal.welch`
    to process each time slice. So, the "modified" method is to use
    the PSD averaging (via welch) for each time slice but then take
    the peaks over all these averages.

    For a pure 'maximax' PSD, just set `timeslice` to ``nperseg/sr``
    and `tsoverlap` to 0.5 (assuming 50% overlap is desired).
    Conversely, for a pure Welch periodogram, just set the `timeslice`
    equal to the entire signal (or just use :func:`scipy.signal.welch`
    of course). Usually the desired behavior for :func:`psdmod` is
    somewhere between these two extremes.

    Examples
    --------
    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from pyyeti import psd
        >>> from scipy import signal
        >>> TF = 30  # make a 30 second signal
        >>> spec = np.array([[20, 1], [50, 1]])
        >>> sig, sr, t = psd.psd2time(spec, ppc=10, fstart=20,
        ...                           fstop=50, df=1/TF,
        ...                           winends=dict(portion=10),
        ...                           gettime=True)
        >>> f, p = signal.welch(sig, sr, nperseg=sr)
        >>> f2, p2 = psd.psdmod(sig, sr, nperseg=sr, timeslice=4,
        ...                     tsoverlap=0.5)
        >>> f3, p3 = psd.psdmod(sig, sr, nperseg=sr)
        >>> spec = spec.T
        >>> fig = plt.figure('Example')
        >>> fig.clf()
        >>> _ = plt.subplot(211)
        >>> _ = plt.plot(t, sig)
        >>> _ = plt.title(r'Input Signal - Specification Level = '
        ...               '1.0 $g^{2}$/Hz')
        >>> _ = plt.xlabel('Time (sec)')
        >>> _ = plt.ylabel('Acceleration (g)')
        >>> _ = plt.subplot(212)
        >>> _ = plt.plot(*spec, 'k-', lw=1.5, label='Spec')
        >>> _ = plt.plot(f, p, label='Welch PSD')
        >>> _ = plt.plot(f2, p2, label='PSDmod')
        >>> _ = plt.plot(f3, p3, label='Maximax')
        >>> _ = plt.legend(loc='best')
        >>> _ = plt.xlim(20, 50)
        >>> _ = plt.title('PSD')
        >>> _ = plt.xlabel('Frequency (Hz)')
        >>> _ = plt.ylabel(r'PSD ($g^2$/Hz)')
        >>> _ = plt.tight_layout()
    """
    if nperseg is None:
        nperseg = int(sr / 5)
    ntimeslice = dsp._proc_timeslice(timeslice, sr, sig.size)[0]
    if nperseg > ntimeslice:
        raise ValueError(
            "`nperseg` too big for current `timeslice` setting;"
            " either decrease `nperseg` or increase `timeslice`"
        )
    welch_inputs = dict(fs=sr, nperseg=nperseg, **kwargs)
    pmap, t, f = dsp.waterfall(
        sig,
        sr,
        timeslice,
        tsoverlap,
        signal.welch,
        which=1,
        freq=0,
        kwargs=welch_inputs,
    )
    p = pmap.max(axis=1)
    if getmap:
        return f, p, pmap, t
    return f, p
